#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Givens rotations routines."""
import numpy

from openfermion.config import EQ_TOLERANCE


def givens_matrix_elements(a, b, which='left'):
    """Compute the matrix elements of the Givens rotation that zeroes out one
    of two row entries.

    If `which='left'` then returns a matrix G such that

        G * [a  b]^T= [0  r]^T

    otherwise, returns a matrix G such that

        G * [a  b]^T= [r  0]^T

    where r is a complex number.

    Args:
        a(complex or float): A complex number representing the upper row entry
        b(complex or float): A complex number representing the lower row entry
        which(string): Either 'left' or 'right', indicating whether to
            zero out the left element (first argument) or right element
            (second argument). Default is `left`.
    Returns:
        G(ndarray): A 2 x 2 numpy array representing the matrix G.
            The numbers in the first column of G are real.
    """
    # Handle case that a is zero
    if abs(a) < EQ_TOLERANCE:
        cosine = 1.
        sine = 0.
        phase = 1.
    # Handle case that b is zero and a is nonzero
    elif abs(b) < EQ_TOLERANCE:
        cosine = 0.
        sine = 1.
        phase = 1.
    # Handle case that a and b are both nonzero
    else:
        denominator = numpy.sqrt(abs(a)**2 + abs(b)**2)
        cosine = abs(b) / denominator
        sine = abs(a) / denominator
        sign_b = b / abs(b)
        sign_a = a / abs(a)
        phase = sign_a * sign_b.conjugate()
        # If phase is a real number, convert it to a float
        if numpy.isreal(phase):
            phase = numpy.real(phase)

    # Construct matrix and return
    if which == 'left':
        # We want to zero out a
        if (abs(numpy.imag(a)) < EQ_TOLERANCE and
                abs(numpy.imag(b)) < EQ_TOLERANCE):
            # a and b are real, so return a standard rotation matrix
            givens_rotation = numpy.array([[cosine, -phase * sine],
                                           [phase * sine, cosine]])
        else:
            givens_rotation = numpy.array([[cosine, -phase * sine],
                                           [sine, phase * cosine]])
    elif which == 'right':
        # We want to zero out b
        if (abs(numpy.imag(a)) < EQ_TOLERANCE and
                abs(numpy.imag(b)) < EQ_TOLERANCE):
            # a and b are real, so return a standard rotation matrix
            givens_rotation = numpy.array([[sine, phase * cosine],
                                           [-phase * cosine, sine]])
        else:
            givens_rotation = numpy.array([[sine, phase * cosine],
                                           [cosine, -phase * sine]])
    else:
        raise ValueError('"which" must be equal to "left" or "right".')
    return givens_rotation


def givens_rotate(operator, givens_rotation, i, j, which='row'):
    """Apply a Givens rotation to coordinates i and j of an operator."""
    if which == 'row':
        # Rotate rows i and j
        row_i = operator[i].copy()
        row_j = operator[j].copy()
        operator[i] = (givens_rotation[0, 0] * row_i +
                       givens_rotation[0, 1] * row_j)
        operator[j] = (givens_rotation[1, 0] * row_i +
                       givens_rotation[1, 1] * row_j)
    elif which == 'col':
        # Rotate columns i and j
        col_i = operator[:, i].copy()
        col_j = operator[:, j].copy()
        operator[:, i] = (givens_rotation[0, 0] * col_i +
                          givens_rotation[0, 1].conj() * col_j)
        operator[:, j] = (givens_rotation[1, 0] * col_i +
                          givens_rotation[1, 1].conj() * col_j)
    else:
        raise ValueError('"which" must be equal to "row" or "col".')


def double_givens_rotate(operator, givens_rotation, i, j, which='row'):
    """Apply a double Givens rotation.

    Applies a Givens rotation to coordinates i and j and the the conjugate
    Givens rotation to coordinates n + i and n + j, where
    n = dim(operator) / 2. dim(operator) must be even.
    """
    m, p = operator.shape

    if which == 'row':
        if m % 2 != 0:
            raise ValueError('To apply a double Givens rotation on rows, '
                             'the number of rows must be even.')
        n = m // 2
        # Rotate rows i and j
        givens_rotate(operator[:n], givens_rotation, i, j, which='row')
        # Rotate rows n + i and n + j
        givens_rotate(operator[n:], givens_rotation.conj(), i, j, which='row')
    elif which == 'col':
        if p % 2 != 0:
            raise ValueError('To apply a double Givens rotation on columns, '
                             'the number of columns must be even.')
        n = p // 2
        # Rotate columns i and j
        givens_rotate(operator[:, :n], givens_rotation, i, j, which='col')
        # Rotate cols n + i and n + j
        givens_rotate(operator[:, n:],
                      givens_rotation.conj(),
                      i,
                      j,
                      which='col')
    else:
        raise ValueError('"which" must be equal to "row" or "col".')


def givens_decomposition_square(unitary_matrix, always_insert=False):
    r"""Decompose a square matrix into a sequence of Givens rotations.

    The input is a square :math:`n \times n` matrix :math:`Q`.
    :math:`Q` can be decomposed as follows:

    .. math::

        Q = DU

    where :math:`U` is unitary and :math:`D` is diagonal.
    Furthermore, we can decompose :math:`U` as

    .. math::

        U = G_k ... G_1

    where :math:`G_1, \ldots, G_k` are complex Givens rotations.
    A Givens rotation is a rotation within the two-dimensional subspace
    spanned by two coordinate axes. Within the two relevant coordinate
    axes, a Givens rotation has the form

    .. math::

        \begin{pmatrix}
            \cos(\theta) & -e^{i \varphi} \sin(\theta) \\
            \sin(\theta) &     e^{i \varphi} \cos(\theta)
        \end{pmatrix}.

    Args:
        unitary_matrix: A numpy array with orthonormal rows,
            representing the matrix Q.

    Returns
    -------
        decomposition (list[tuple]):
            A list of tuples of objects describing Givens
            rotations. The list looks like [(G_1, ), (G_2, G_3), ... ].
            The Givens rotations within a tuple can be implemented in parallel.
            The description of a Givens rotation is itself a tuple of the
            form :math:`(i, j, \theta, \varphi)`, which represents a
            Givens rotation of coordinates
            :math:`i` and :math:`j` by angles :math:`\theta` and
            :math:`\varphi`.
        diagonal (ndarray):
            A list of the nonzero entries of :math:`D`.
    """
    current_matrix = numpy.copy(unitary_matrix)
    n = current_matrix.shape[0]

    decomposition = []

    for k in range(2 * (n - 1) - 1):
        # Initialize the list of parallel operations to perform
        # in this iteration
        parallel_ops = []

        # Get the (row, column) indices of elements to zero out in parallel.
        if k < n - 1:
            start_row = 0
            start_column = n - 1 - k
        else:
            start_row = k - (n - 2)
            start_column = k - (n - 3)
        column_indices = range(start_column, n, 2)
        row_indices = range(start_row, start_row + len(column_indices))
        indices_to_zero_out = zip(row_indices, column_indices)

        for i, j in indices_to_zero_out:
            # Compute the Givens rotation to zero out the (i, j) element,
            # if needed
            right_element = current_matrix[i, j].conj()
            if always_insert or abs(right_element) > EQ_TOLERANCE:
                # We actually need to perform a Givens rotation
                left_element = current_matrix[i, j - 1].conj()
                givens_rotation = givens_matrix_elements(left_element,
                                                         right_element,
                                                         which='right')

                # Add the parameters to the list
                theta = numpy.arcsin(numpy.real(givens_rotation[1, 0]))
                phi = numpy.angle(givens_rotation[1, 1])
                parallel_ops.append((j - 1, j, theta, phi))

                # Update the matrix
                givens_rotate(current_matrix,
                              givens_rotation,
                              j - 1,
                              j,
                              which='col')

        # If the current list of parallel operations is not empty,
        # append it to the list,
        if parallel_ops:
            decomposition.append(tuple(parallel_ops))

    # Get the diagonal entries
    diagonal = current_matrix[range(n), range(n)]

    return decomposition, diagonal


def givens_decomposition(unitary_rows, always_insert=False):
    r"""Decompose a matrix into a sequence of Givens rotations.

    The input is an :math:`m \times n` matrix :math:`Q` with :math:`m \leq n`.
    The rows of :math:`Q` are orthonormal.
    :math:`Q` can be decomposed as follows:

    .. math::

        V Q U^\dagger = D

    where :math:`V` and :math:`U` are unitary matrices, and :math:`D`
    is an :math:`m \times n` matrix with the
    first :math:`m` columns forming a diagonal matrix and the rest of the
    columns being zero. Furthermore, we can decompose :math:`U` as

    .. math::

        U = G_k ... G_1

    where :math:`G_1, \ldots, G_k` are complex Givens rotations.
    A Givens rotation is a rotation within the two-dimensional subspace
    spanned by two coordinate axes. Within the two relevant coordinate
    axes, a Givens rotation has the form

    .. math::

        \begin{pmatrix}
            \cos(\theta) & -e^{i \varphi} \sin(\theta) \\
            \sin(\theta) &     e^{i \varphi} \cos(\theta)
        \end{pmatrix}.

    Args:
        unitary_rows: A numpy array or matrix with orthonormal rows,
            representing the matrix Q.

    Returns
    -------
        givens_rotations (list[tuple]):
            A list of tuples of objects describing Givens
            rotations. The list looks like [(G_1, ), (G_2, G_3), ... ].
            The Givens rotations within a tuple can be implemented in parallel.
            The description of a Givens rotation is itself a tuple of the
            form :math:`(i, j, \theta, \varphi)`, which represents a
            Givens rotation of coordinates
            :math:`i` and :math:`j` by angles :math:`\theta` and
            :math:`\varphi`.
        left_unitary (ndarray):
            An :math:`m \times m` numpy array representing the matrix
            :math:`V`.
        diagonal (ndarray):
            A list of the nonzero entries of :math:`D`.
    """
    current_matrix = numpy.copy(unitary_rows)
    m, n = current_matrix.shape

    # Check that m <= n
    if m > n:
        raise ValueError('The input m x n matrix must have m <= n')

    # Compute left_unitary using Givens rotations
    left_unitary = numpy.eye(m, dtype=complex)
    for k in reversed(range(n - m + 1, n)):
        # Zero out entries in column k
        for l in range(m - n + k):
            # Zero out entry in row l if needed
            if abs(current_matrix[l, k]) > EQ_TOLERANCE:
                givens_rotation = givens_matrix_elements(
                    current_matrix[l, k], current_matrix[l + 1, k])
                # Apply Givens rotation
                givens_rotate(current_matrix, givens_rotation, l, l + 1)
                givens_rotate(left_unitary, givens_rotation, l, l + 1)

    # Compute the decomposition of current_matrix into Givens rotations
    givens_rotations = []
    # If m = n (the matrix is square) then we don't need to perform any
    # Givens rotations!
    if m != n:
        # Get the maximum number of simultaneous rotations that
        # will be performed
        max_simul_rotations = min(m, n - m)
        # There are n - 1 iterations (the circuit depth is n - 1)
        for k in range(n - 1):
            # Get the (row, column) indices of elements to zero out in
            # parallel.
            if k < max_simul_rotations - 1:
                # There are k + 1 elements to zero out
                start_row = 0
                end_row = k + 1
                start_column = n - m - k
                end_column = start_column + 2 * (k + 1)
            elif k > n - 1 - max_simul_rotations:
                # There are n - 1 - k elements to zero out
                start_row = m - (n - 1 - k)
                end_row = m
                start_column = m - (n - 1 - k) + 1
                end_column = start_column + 2 * (n - 1 - k)
            else:
                # There are max_simul_rotations elements to zero out
                if max_simul_rotations == m:
                    start_row = 0
                    end_row = m
                    start_column = n - m - k
                    end_column = start_column + 2 * m
                else:
                    start_row = k + 1 - max_simul_rotations
                    end_row = k + 1
                    start_column = k + 1 - max_simul_rotations + 1
                    end_column = start_column + 2 * max_simul_rotations

            row_indices = range(start_row, end_row)
            column_indices = range(start_column, end_column, 2)
            indices_to_zero_out = zip(row_indices, column_indices)

            parallel_rotations = []
            for i, j in indices_to_zero_out:
                # Compute the Givens rotation to zero out the (i, j) element,
                # if needed
                right_element = current_matrix[i, j].conj()
                if always_insert or abs(right_element) > EQ_TOLERANCE:
                    # We actually need to perform a Givens rotation
                    left_element = current_matrix[i, j - 1].conj()
                    givens_rotation = givens_matrix_elements(left_element,
                                                             right_element,
                                                             which='right')

                    # Add the parameters to the list
                    theta = numpy.arcsin(numpy.real(givens_rotation[1, 0]))
                    phi = numpy.angle(givens_rotation[1, 1])
                    parallel_rotations.append((j - 1, j, theta, phi))

                    # Update the matrix
                    givens_rotate(current_matrix,
                                  givens_rotation,
                                  j - 1,
                                  j,
                                  which='col')

            # If the current list of parallel operations is not empty,
            # append it to the list,
            if parallel_rotations:
                givens_rotations.append(tuple(parallel_rotations))

    # Get the diagonal entries
    diagonal = current_matrix.diagonal()

    return givens_rotations, left_unitary, diagonal


def fermionic_gaussian_decomposition(unitary_rows):
    r"""Decompose a matrix into a sequence of Givens rotations and
    particle-hole transformations on the last fermionic mode.

    The input is an :math:`N \times 2N` matrix :math:`W` with orthonormal
    rows. Furthermore, :math:`W` must have the block form

    .. math::

        W = ( W_1 \hspace{4pt} W_2 )

    where :math:`W_1` and :math:`W_2` satisfy

    .. math::

        W_1  W_1^\dagger + W_2  W_2^\dagger &= I

        W_1  W_2^T + W_2  W_1^T &= 0.

    Then :math:`W` can be decomposed as

    .. math::

        V  W  U^\dagger = ( 0 \hspace{6pt} D )

    where :math:`V` and :math:`U` are unitary matrices and :math:`D`
    is a diagonal unitary matrix. Furthermore, :math:`U` can be decomposed
    as follows:

    .. math::

        U = B G_{k} \cdots B G_3 G_2 B G_1 B,

    where each :math:`G_i` is a Givens rotation, and :math:`B` represents
    swapping the :math:`N`-th column with the :math:`2N`-th column,
    which corresponds to a particle-hole transformation
    on the last fermionic mode. This particle-hole transformation maps
    :math:`a^\dagger_N` to :math:`a_N` and vice versa, while leaving the
    other fermionic ladder operators invariant.

    The decomposition of :math:`U` is returned as a list of tuples of objects
    describing rotations and particle-hole transformations. The list looks
    something like [('pht', ), (G_1, ), ('pht', G_2), ... ].
    The objects within a tuple are either the string 'pht', which indicates
    a particle-hole transformation on the last fermionic mode, or a tuple
    of the form :math:`(i, j, \theta, \varphi)`, which indicates a
    Givens rotation of rows :math:`i` and :math:`j` by angles
    :math:`\theta` and :math:`\varphi`.

    The matrix :math:`V^T D^*` can also be decomposed as a sequence of
    Givens rotations. This decomposition is needed for a circuit that
    prepares an excited state.

    Args:
        unitary_rows(ndarray): A matrix with orthonormal rows and
            additional structure described above.

    Returns
    -------
        decomposition (list[tuple]):
            The decomposition of :math:`U`.
        left_decomposition (list[tuple]):
            The decomposition of :math:`V^T D^*`.
        diagonal (ndarray):
            A list of the nonzero entries of :math:`D`.
        left_diagonal (ndarray):
            A list of the nonzero entries left from the decomposition
            of :math:`V^T D^*`.
    """
    current_matrix = numpy.copy(unitary_rows)
    n, p = current_matrix.shape

    # Check that p = 2 * n
    if p != 2 * n:
        raise ValueError('The input matrix must have twice as many columns '
                         'as rows.')

    # Check that left and right parts of unitary_rows satisfy the constraints
    # necessary for the transformed fermionic operators to satisfy
    # the fermionic anticommutation relations
    left_part = unitary_rows[:, :n]
    right_part = unitary_rows[:, n:]
    constraint_matrix_1 = (left_part.dot(left_part.T.conj()) +
                           right_part.dot(right_part.T.conj()))
    constraint_matrix_2 = (left_part.dot(right_part.T) +
                           right_part.dot(left_part.T))
    discrepancy_1 = numpy.amax(abs(constraint_matrix_1 - numpy.eye(n)))
    discrepancy_2 = numpy.amax(abs(constraint_matrix_2))
    if discrepancy_1 > EQ_TOLERANCE or discrepancy_2 > EQ_TOLERANCE:
        raise ValueError('The input matrix does not satisfy the constraints '
                         'necessary for a proper transformation of the '
                         'fermionic ladder operators.')

    # Compute left_unitary using Givens rotations
    left_unitary = numpy.eye(n, dtype=complex)
    for k in range(n - 1):
        # Zero out entries in column k
        for l in range(n - 1 - k):
            # Zero out entry in row l if needed
            if abs(current_matrix[l, k]) > EQ_TOLERANCE:
                givens_rotation = givens_matrix_elements(
                    current_matrix[l, k], current_matrix[l + 1, k])
                # Apply Givens rotation
                givens_rotate(current_matrix, givens_rotation, l, l + 1)
                givens_rotate(left_unitary, givens_rotation, l, l + 1)

    # Initialize list to store decomposition of current_matrix
    decomposition = []
    # There are 2 * n - 1 iterations (that is the circuit depth)
    for k in range(2 * n - 1):
        # Initialize the list of parallel operations to perform
        # in this iteration
        parallel_ops = []

        # Perform a particle-hole transformation if necessary
        if k % 2 == 0 and abs(current_matrix[k // 2, n - 1]) > EQ_TOLERANCE:
            parallel_ops.append('pht')
            swap_columns(current_matrix, n - 1, 2 * n - 1)

        # Get the (row, column) indices of elements to zero out in parallel.
        if k < n:
            end_row = k
            end_column = n - 1 - k
        else:
            end_row = n - 1
            end_column = k - (n - 1)
        column_indices = range(end_column, n - 1, 2)
        row_indices = range(end_row, end_row - len(column_indices), -1)
        indices_to_zero_out = zip(row_indices, column_indices)

        for i, j in indices_to_zero_out:
            # Compute the Givens rotation to zero out the (i, j) element,
            # if needed
            left_element = current_matrix[i, j].conj()
            if abs(left_element) > EQ_TOLERANCE:
                # We actually need to perform a Givens rotation
                right_element = current_matrix[i, j + 1].conj()
                givens_rotation = givens_matrix_elements(
                    left_element, right_element)

                # Add the parameters to the list
                theta = numpy.arcsin(numpy.real(givens_rotation[1, 0]))
                phi = numpy.angle(givens_rotation[1, 1])
                parallel_ops.append((j, j + 1, theta, phi))

                # Update the matrix
                double_givens_rotate(current_matrix,
                                     givens_rotation,
                                     j,
                                     j + 1,
                                     which='col')

        # If the current list of parallel operations is not empty,
        # append it to the list,
        if parallel_ops:
            decomposition.append(tuple(parallel_ops))

    # Get the diagonal entries
    diagonal = current_matrix[range(n), range(n, 2 * n)]

    # Compute the decomposition of left_unitary^T * diagonal^*
    current_matrix = left_unitary.T
    for k in range(n):
        current_matrix[:, k] *= diagonal[k].conj()

    left_decomposition, left_diagonal = givens_decomposition_square(
        current_matrix)

    return decomposition, left_decomposition, diagonal, left_diagonal


def swap_rows(M, i, j):
    """Swap rows i and j of matrix M."""
    if len(M.shape) == 1:
        M[i], M[j] = M[j], M[i]
    else:
        row_i = M[i, :].copy()
        row_j = M[j, :].copy()
        M[i, :], M[j, :] = row_j, row_i


def swap_columns(M, i, j):
    """Swap columns i and j of matrix M."""
    if len(M.shape) == 1:
        M[i], M[j] = M[j], M[i]
    else:
        column_i = M[:, i].copy()
        column_j = M[:, j].copy()
        M[:, i], M[:, j] = column_j, column_i



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from .equality_constraint_projection import (
    apply_constraints,
    constraint_matrix,
    linearize_term,
    unlinearize_term,
    operator_to_vector,
    vector_to_operator,
)

from .qubit_partitioning import (
    binary_partition_iterator,
    group_into_tensor_product_basis_sets,
    partition_iterator,
    pauli_string_iterator,
)

from .fermion_partitioning import (
    pair_within,
    pair_between,
    pair_within_simultaneously,
    pair_within_simultaneously_binned,
    pair_within_simultaneously_symmetric,
)

from .get_interaction_rdm import get_interaction_rdm

from .rdm_equality_constraints import (
    one_body_fermion_constraints,
    two_body_fermion_constraints,
)

from .vpe_estimators import (
    PhaseFitEstimator,
    get_phase_function,
)



"""Code to generate the eigenvalue problem for the ERPA equations"""
from typing import Dict, Tuple, Union
from itertools import product
import numpy
from numpy import einsum
import scipy
from openfermion.utils.rdm_mapping_functions import kronecker_delta as kdelta
from openfermion.utils.rdm_mapping_functions import map_two_pdm_to_one_pdm


def erpa_eom_hamiltonian(h_ijkl: numpy.ndarray, tpdm: numpy.ndarray, p: int,
                         q: int, r: int, s: int) -> Union[float, complex]:
    """
    Evaluate sum_{a,b,c,d}h_{a, b, d, c}<psi[p^ q, [a^ b^ c d, r^ s]]psi>

    Args:
        h_ijkl: two-body integral tensors of full reduced Hamiltonian
        tpdm: two-RDM
        p: left creation op index
        q: left annihilation op index
        r: right creation op index
        s: right annihilation op index
    Returns:
        float or complex number
    """
    h_mat = 0
    #  (   1.00000) h_ijkl(q,s,a,b) cre(p) cre(r) des(a) des(b)
    h_mat += 1.0 * einsum('ab,ab', h_ijkl[q, s, :, :], tpdm[p, r, :, :])

    #  (  -1.00000) h_ijkl(q,a,r,b) cre(p) cre(a) des(s) des(b)
    h_mat += -1.0 * einsum('ab,ab', h_ijkl[q, :, r, :], tpdm[p, :, s, :])

    #  (   1.00000) h_ijkl(q,a,b,r) cre(p) cre(a) des(s) des(b)
    h_mat += 1.0 * einsum('ab,ab', h_ijkl[q, :, :, r], tpdm[p, :, s, :])

    #  (  -1.00000) h_ijkl(s,q,a,b) cre(p) cre(r) des(a) des(b)
    h_mat += -1.0 * einsum('ab,ab', h_ijkl[s, q, :, :], tpdm[p, r, :, :])

    #  (  -1.00000) h_ijkl(s,a,p,b) cre(r) cre(a) des(q) des(b)
    h_mat += -1.0 * einsum('ab,ab', h_ijkl[s, :, p, :], tpdm[r, :, q, :])

    #  (   1.00000) h_ijkl(s,a,b,p) cre(r) cre(a) des(q) des(b)
    h_mat += 1.0 * einsum('ab,ab', h_ijkl[s, :, :, p], tpdm[r, :, q, :])

    #  (   1.00000) h_ijkl(a,q,r,b) cre(p) cre(a) des(s) des(b)
    h_mat += 1.0 * einsum('ab,ab', h_ijkl[:, q, r, :], tpdm[p, :, s, :])

    #  (  -1.00000) h_ijkl(a,q,b,r) cre(p) cre(a) des(s) des(b)
    h_mat += -1.0 * einsum('ab,ab', h_ijkl[:, q, :, r], tpdm[p, :, s, :])

    #  (   1.00000) h_ijkl(a,s,p,b) cre(r) cre(a) des(q) des(b)
    h_mat += 1.0 * einsum('ab,ab', h_ijkl[:, s, p, :], tpdm[r, :, q, :])

    #  (  -1.00000) h_ijkl(a,s,b,p) cre(r) cre(a) des(q) des(b)
    h_mat += -1.0 * einsum('ab,ab', h_ijkl[:, s, :, p], tpdm[r, :, q, :])

    #  (   1.00000) h_ijkl(a,b,p,r) cre(a) cre(b) des(q) des(s)
    h_mat += 1.0 * einsum('ab,ab', h_ijkl[:, :, p, r], tpdm[:, :, q, s])

    #  (  -1.00000) h_ijkl(a,b,r,p) cre(a) cre(b) des(q) des(s)
    h_mat += -1.0 * einsum('ab,ab', h_ijkl[:, :, r, p], tpdm[:, :, q, s])

    if q == r:
        #  (   1.00000) h_ijkl(s,a,b,c) kdelta(q,r) cre(p) cre(a) des(b) des(c)
        h_mat += 1.0 * einsum('abc,abc', h_ijkl[s, :, :, :], tpdm[p, :, :, :])

        #  (  -1.00000) h_ijkl(a,s,b,c) kdelta(q,r) cre(p) cre(a) des(b) des(c)
        h_mat += -1.0 * einsum('abc,abc', h_ijkl[:, s, :, :], tpdm[p, :, :, :])

    if p == s:
        #  (   1.00000) h_ijkl(a,b,r,c) kdelta(p,s) cre(a) cre(b) des(q) des(c)
        h_mat += 1.0 * einsum('abc,abc', h_ijkl[:, :, r, :], tpdm[:, :, q, :])

        #  (  -1.00000) h_ijkl(a,b,c,r) kdelta(p,s) cre(a) cre(b) des(q) des(c)
        h_mat += -1.0 * einsum('abc,abc', h_ijkl[:, :, :, r], tpdm[:, :, q, :])

    return h_mat


def singlet_erpa(tpdm: numpy.ndarray, h_ijkl: numpy.ndarray) \
        -> Tuple[numpy.ndarray, numpy.ndarray, Dict]:
    """
    Generate the singlet ERPA equations

    [ea + eb, [H, sa, sb]] = [ea, [H, sa]]

    The erpa equations are solved with scipy.linalg.eig which calls
    lapack's geev

    Args:
        tpdm: 2-RDM tensor normal OpenFermion format
        h_ijkl: reduced Hamiltonian tensor
    Returns:
        Tuple of the erpa system.
    """
    permuted_hijkl = numpy.einsum('ijlk', h_ijkl)
    roots = numpy.array(numpy.roots([1, -1, -einsum('ijji', tpdm)]))
    n_electrons = roots[numpy.where(roots > 0)[0]].real
    opdm = map_two_pdm_to_one_pdm(tpdm, n_electrons)
    dim = tpdm.shape[0] // 2  # dim = num spatial orbitals
    full_basis = {}  # erpa basis.  A, B basis in RPA language
    cnt = 0
    for p, q in product(range(dim), repeat=2):
        if p < q:
            full_basis[(p, q)] = cnt
            full_basis[(q, p)] = cnt + dim * (dim - 1) // 2
            cnt += 1

    erpa_mat = numpy.zeros((len(full_basis), len(full_basis)))
    metric_mat = numpy.zeros((len(full_basis), len(full_basis)))
    for rkey, ridx in full_basis.items():
        p, q = rkey
        for ckey, cidx in full_basis.items():
            r, s = ckey
            for sigma, tau in product([0, 1], repeat=2):
                erpa_mat[ridx, cidx] += 0.5 * erpa_eom_hamiltonian(
                    permuted_hijkl, tpdm, 2 * q + sigma, 2 * p + sigma,
                    2 * r + tau, 2 * s + tau).real
                metric_mat[ridx, cidx] += 0.5 * (
                    opdm[2 * q + sigma, 2 * s + tau] *
                    kdelta(2 * r + tau, 2 * p + sigma) -
                    opdm[2 * p + sigma, 2 * r + tau] *
                    kdelta(2 * q + sigma, 2 * s + tau)).real

    # The metric is hermetian and can be diagonalized
    # this allows us to project into the non-zero eigenvalue space
    ws, vs = numpy.linalg.eigh(metric_mat)
    non_zero_idx = numpy.where(numpy.abs(ws) > 1.0E-8)[0]
    left_mat = vs[:, non_zero_idx].T @ erpa_mat @ vs[:, non_zero_idx]
    right_mat = vs[:, non_zero_idx].T @ metric_mat @ vs[:, non_zero_idx]

    # solve the matrix pencil using ddgev in lapack
    w, v = scipy.linalg.eig(left_mat, right_mat)

    # the spectrum is symmetric (-w, w)
    # find the positive spectrum eigensystem and return
    real_eig_idx = numpy.where(numpy.abs(w.imag) < 1.0E-8)[0]
    real_eigs = w[real_eig_idx]
    real_eig_vecs = v[:, real_eig_idx]
    reverse_projected_eig_vecs = vs[:, non_zero_idx] @ real_eig_vecs
    pos_indices = numpy.where(real_eigs > 0)[0]
    pos_indices = pos_indices[numpy.argsort(real_eigs[pos_indices])]
    return real_eigs[pos_indices], reverse_projected_eig_vecs[:, pos_indices], \
           full_basis



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Class and functions to store interaction operators."""
import itertools

import numpy

from openfermion.ops.representations.polynomial_tensor import PolynomialTensor
from openfermion.config import EQ_TOLERANCE


class InteractionOperatorError(Exception):
    pass


class InteractionOperator(PolynomialTensor):
    r"""Class for storing 'interaction operators' which are defined to be
    fermionic operators consisting of one-body and two-body terms which
    conserve particle number and spin. The most common examples of data that
    will use this structure are molecular Hamiltonians. In principle,
    everything stored in this class could also be represented using the more
    general FermionOperator class. However, this class is able to exploit
    specific properties of how fermions interact to enable more numerically
    efficient manipulation of the data. Note that the operators stored in this
    class take the form:

        .. math::

            constant + \sum_{p, q} h_{p, q} a^\dagger_p a_q +
            \sum_{p, q, r, s} h_{p, q, r, s} a^\dagger_p a^\dagger_q a_r a_s.

    Attributes:
        one_body_tensor: The coefficients of the one-body terms
        (:math:`h_{p, q}`). This is an n_qubits x n_qubits
        numpy array of floats.
        two_body_tensor: The coefficients of the two-body terms
            (:math:`h_{p, q, r, s}`).
            This is an n_qubits x n_qubits x n_qubits x
            n_qubits numpy array of floats.
    """

    def __init__(self, constant, one_body_tensor, two_body_tensor):
        """
        Initialize the InteractionOperator class.

        Args:
            constant: A constant term in the operator given as a
                float. For instance, the nuclear repulsion energy.
            one_body_tensor: The coefficients of the one-body terms
                (:math:`h_{p,q}`).
               This is an n_qubits x n_qubits numpy array of floats.
            two_body_tensor: The coefficients of the two-body terms
                (:math:`h_{p, q, r, s}`).
                This is an n_qubits x n_qubits x n_qubits x
                n_qubits numpy array of floats.
        """
        # Make sure nonzero elements are only for normal ordered terms.
        super(InteractionOperator, self).__init__({
            (): constant,
            (1, 0): one_body_tensor,
            (1, 1, 0, 0): two_body_tensor
        })

    @property
    def one_body_tensor(self):
        """The value of the one-body tensor."""
        return self.n_body_tensors[1, 0]

    @one_body_tensor.setter
    def one_body_tensor(self, value):
        """Set the value of the one-body tensor."""
        self.n_body_tensors[1, 0] = value

    @property
    def two_body_tensor(self):
        """The value of the two-body tensor."""
        return self.n_body_tensors[1, 1, 0, 0]

    @two_body_tensor.setter
    def two_body_tensor(self, value):
        """Set the value of the two-body tensor."""
        self.n_body_tensors[1, 1, 0, 0] = value

    def unique_iter(self, complex_valued=False):
        """
        Iterate all terms that are not in the same symmetry group.

        Four point symmetry:
            1. pq = qp.
            2. pqrs = srqp = qpsr = rspq.
        Eight point symmetry:
            1. pq = qp.
            2. pqrs = rqps = psrq = srqp = qpsr = rspq = spqr = qrsp.

        Args:
            complex_valued (bool):
                Whether the operator has complex coefficients.
        Yields:
            tuple[int]
        """
        # Constant.
        if self.constant:
            yield ()

        # One-body terms.
        for p in range(self.n_qubits):
            for q in range(p + 1):
                if self.one_body_tensor[p, q]:
                    yield (p, 1), (q, 0)

        # Two-body terms.
        seen = set()
        for quad in itertools.product(range(self.n_qubits), repeat=4):
            if self.two_body_tensor[quad] and quad not in seen:
                seen |= set(_symmetric_two_body_terms(quad, complex_valued))
                yield tuple(zip(quad, (1, 1, 0, 0)))

    @classmethod
    def zero(cls, n_qubits):
        return cls(0, numpy.zeros((n_qubits,) * 2, dtype=numpy.complex128),
                   numpy.zeros((n_qubits,) * 4, dtype=numpy.complex128))

    def projected(self, indices, exact=False):
        projected_n_body_tensors = self.projected_n_body_tensors(indices, exact)
        return type(self)(*(projected_n_body_tensors[key]
                            for key in [(), (1, 0), (1, 1, 0, 0)]))

    def with_function_applied_elementwise(self, func):
        return type(self)(*(
            func(tensor) for tensor in
            [self.constant, self.one_body_tensor, self.two_body_tensor]))


def _symmetric_two_body_terms(quad, complex_valued):
    p, q, r, s = quad
    yield p, q, r, s
    yield q, p, s, r
    yield s, r, q, p
    yield r, s, p, q
    if not complex_valued:
        yield p, s, r, q
        yield q, r, s, p
        yield s, p, q, r
        yield r, q, p, s


def get_tensors_from_integrals(one_body_integrals, two_body_integrals):
    '''Converts one and two-body integrals into tensor form

    Arguments:
        one_body_integrals [numpy array] -- the one-body integrals
            of the given Hamiltonian
        two_body_integrals [numpy array] -- the two-body integrals
            of the given Hamiltonian
    '''

    n_qubits = 2 * one_body_integrals.shape[0]

    # Initialize Hamiltonian coefficients.
    one_body_coefficients = numpy.zeros((n_qubits, n_qubits))
    two_body_coefficients = numpy.zeros(
        (n_qubits, n_qubits, n_qubits, n_qubits))
    # Loop through integrals.
    for p in range(n_qubits // 2):
        for q in range(n_qubits // 2):

            # Populate 1-body coefficients. Require p and q have same spin.
            one_body_coefficients[2 * p, 2 * q] = one_body_integrals[p, q]
            one_body_coefficients[2 * p + 1, 2 * q +
                                  1] = one_body_integrals[p, q]
            # Continue looping to prepare 2-body coefficients.
            for r in range(n_qubits // 2):
                for s in range(n_qubits // 2):

                    # Mixed spin
                    two_body_coefficients[2 * p, 2 * q + 1, 2 * r + 1, 2 *
                                          s] = (two_body_integrals[p, q, r, s] /
                                                2.)
                    two_body_coefficients[2 * p + 1, 2 * q, 2 * r, 2 * s +
                                          1] = (two_body_integrals[p, q, r, s] /
                                                2.)

                    # Same spin
                    two_body_coefficients[2 * p, 2 * q, 2 * r, 2 *
                                          s] = (two_body_integrals[p, q, r, s] /
                                                2.)
                    two_body_coefficients[2 * p + 1, 2 * q + 1, 2 * r +
                                          1, 2 * s +
                                          1] = (two_body_integrals[p, q, r, s] /
                                                2.)

    # Truncate.
    one_body_coefficients[
        numpy.absolute(one_body_coefficients) < EQ_TOLERANCE] = 0.
    two_body_coefficients[
        numpy.absolute(two_body_coefficients) < EQ_TOLERANCE] = 0.

    return one_body_coefficients, two_body_coefficients


def get_active_space_integrals(one_body_integrals,
                               two_body_integrals,
                               occupied_indices=None,
                               active_indices=None):
    """Restricts a molecule at a spatial orbital level to an active space

    This active space may be defined by a list of active indices and
        doubly occupied indices. Note that one_body_integrals and
        two_body_integrals must be defined
        n an orthonormal basis set.

    Args:
        one_body_integrals: One-body integrals of the target Hamiltonian
        two_body_integrals: Two-body integrals of the target Hamiltonian
        occupied_indices: A list of spatial orbital indices
            indicating which orbitals should be considered doubly occupied.
        active_indices: A list of spatial orbital indices indicating
            which orbitals should be considered active.

    Returns:
        tuple: Tuple with the following entries:

        **core_constant**: Adjustment to constant shift in Hamiltonian
        from integrating out core orbitals

        **one_body_integrals_new**: one-electron integrals over active
        space.

        **two_body_integrals_new**: two-electron integrals over active
        space.
    """
    # Fix data type for a few edge cases
    occupied_indices = [] if occupied_indices is None else occupied_indices
    if (len(active_indices) < 1):
        raise ValueError('Some active indices required for reduction.')

    # Determine core constant
    core_constant = 0.0
    for i in occupied_indices:
        core_constant += 2 * one_body_integrals[i, i]
        for j in occupied_indices:
            core_constant += (2 * two_body_integrals[i, j, j, i] -
                              two_body_integrals[i, j, i, j])

    # Modified one electron integrals
    one_body_integrals_new = numpy.copy(one_body_integrals)
    for u in active_indices:
        for v in active_indices:
            for i in occupied_indices:
                one_body_integrals_new[u, v] += (
                    2 * two_body_integrals[i, u, v, i] -
                    two_body_integrals[i, u, i, v])

    # Restrict integral ranges and change M appropriately
    return (core_constant,
            one_body_integrals_new[numpy.ix_(active_indices, active_indices)],
            two_body_integrals[numpy.ix_(active_indices, active_indices,
                                         active_indices, active_indices)])



import itertools
import numpy

from openfermion.ops.operators import FermionOperator
from openfermion.ops.representations import InteractionRDM

from openfermion.utils.operator_utils import count_qubits
from openfermion.transforms.opconversions.conversions import check_no_sympy


def get_interaction_rdm(qubit_operator, n_qubits=None):
    """Build an InteractionRDM from measured qubit operators.

    Returns: An InteractionRDM object.
    """

    check_no_sympy(qubit_operator)

    # Avoid circular import.
    from openfermion.transforms import jordan_wigner
    if n_qubits is None:
        n_qubits = count_qubits(qubit_operator)
    one_rdm = numpy.zeros((n_qubits,) * 2, dtype=complex)
    two_rdm = numpy.zeros((n_qubits,) * 4, dtype=complex)

    # One-RDM.
    for i, j in itertools.product(range(n_qubits), repeat=2):
        transformed_operator = jordan_wigner(FermionOperator(((i, 1), (j, 0))))
        for term, coefficient in transformed_operator.terms.items():
            if term in qubit_operator.terms:
                one_rdm[i, j] += coefficient * qubit_operator.terms[term]

    # Two-RDM.
    for i, j, k, l in itertools.product(range(n_qubits), repeat=4):
        transformed_operator = jordan_wigner(
            FermionOperator(((i, 1), (j, 1), (k, 0), (l, 0))))
        for term, coefficient in transformed_operator.terms.items():
            if term in qubit_operator.terms:
                two_rdm[i, j, k, l] += coefficient * qubit_operator.terms[term]

    return InteractionRDM(one_rdm, two_rdm)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from .opconversions import (
    commutator_ordered_diagonal_coulomb_with_two_body_operator,
    chemist_ordered,
    normal_ordered,
    normal_ordered_ladder_term,
    normal_ordered_quad_term,
    reorder,
    linearize_decoder,
    checksum_code,
    bravyi_kitaev_code,
    jordan_wigner_code,
    parity_code,
    weight_one_binary_addressing_code,
    weight_one_segment_code,
    weight_two_segment_code,
    interleaved_code,
    binary_code_transform,
    bravyi_kitaev_fast,
    bravyi_kitaev_fast_interaction_op,
    bravyi_kitaev_fast_edge_matrix,
    bravyi_kitaev,
    bravyi_kitaev_tree,
    get_fermion_operator,
    get_boson_operator,
    get_majorana_operator,
    get_quad_operator,
    check_no_sympy,
    FenwickNode,
    FenwickTree,
    jordan_wigner,
    jordan_wigner_one_body,
    jordan_wigner_two_body,
    qubit_operator_to_pauli_sum,
    reverse_jordan_wigner,
    symmetry_conserving_bravyi_kitaev,
    verstraete_cirac_2d_square,
    vertical_edges_snake,
)

from .repconversions import (
    get_interaction_operator,
    get_diagonal_coulomb_hamiltonian,
    get_molecular_data,
    get_quadratic_hamiltonian,
    fourier_transform,
    inverse_fourier_transform,
    freeze_orbitals,
    prune_unused_indices,
    project_onto_sector,
    projection_error,
    rotate_qubit_by_pauli,
    StabilizerError,
    check_commuting_stabilizers,
    check_stabilizer_linearity,
    reduce_number_of_terms,
    taper_off_qubits,
    fix_single_term,
    mccoy,
    weyl_polynomial_quantization,
    symmetric_ordering,
)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
'''Tools to assist in circuit/primitive design'''

from typing import List

import numpy
from scipy import linalg
import cirq
import openfermion
from openfermion.config import EQ_TOLERANCE


def validate_trotterized_evolution(circuit: cirq.Circuit,
                                   ops: List['openfermion.QubitOperator'],
                                   qubits: List['cirq.Qid']):
    r'''Checks whether a circuit implements Trotterized evolution

    Takes a circuit that is supposed to implement evolution of the
    form:

    $$\prod_j \exp[iO_j]$$

    and checks whether the implemented unitary is applied.
    Ignores any global phases as part of the implementation of the unitary, as
    these are not kept consistent in cirq (and unphysical).

    Arguments:
        circuit: 'cirq.Circuit' {[type]} -- circuit to be checked
        ops {List['openfermion.QubitOperator']} -- list of operators $O_j$
            in application order (i.e. ops[0] is the first operator to be
            applied).
        qubits {List['cirq.Qid']} -- list of qubits in circuit in index order
    '''

    n_qubits = len(qubits)
    hs_dim = 2**n_qubits
    qubit_op = ops[0]
    op_matrix = openfermion.get_sparse_operator(qubit_op, n_qubits=n_qubits)
    target_unitary = linalg.expm(1j * op_matrix)
    for qubit_op in ops[1:]:
        op_matrix = openfermion.get_sparse_operator(qubit_op, n_qubits=n_qubits)
        op_unitary = linalg.expm(1j * op_matrix)
        target_unitary = op_unitary.dot(target_unitary)

    actual_unitary = circuit.unitary(qubit_order=qubits)
    overlap = float(
        numpy.abs(numpy.trace(target_unitary.dot(
            actual_unitary.conj().T)))) / hs_dim
    if 1 - overlap > EQ_TOLERANCE:
        return False
    return True



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
""" Module to remove two qubits from the problem space using conservation
    of electron number and conservation of electron spin. As described in
    arXiv:1701.08213 and Phys. Rev. X 6, 031007.
"""

from openfermion.ops.operators import FermionOperator
from openfermion.transforms.opconversions import bravyi_kitaev_tree, reorder
from openfermion.transforms.repconversions import prune_unused_indices
from openfermion.utils.indexing import up_then_down


def symmetry_conserving_bravyi_kitaev(fermion_hamiltonian, active_orbitals,
                                      active_fermions):
    """ Returns the qubit Hamiltonian for the fermionic Hamiltonian
        supplied, with two qubits removed using conservation of electron
        spin and number, as described in arXiv:1701.08213.

        Args:
            fermion_hamiltonian: A fermionic hamiltonian obtained
                                 using OpenFermion. An instance
                                 of the FermionOperator class.

            active_orbitals: Int type object. The number of active orbitals
                             being considered for the system.

            active_fermions: Int type object. The number of active fermions
                              being considered for the system (note, this
                              is less than the number of electrons in a
                              molecule if some orbitals have been assumed
                              filled).
        Returns:
                qubit_hamiltonian: The qubit Hamiltonian corresponding to
                                   the supplied fermionic Hamiltonian, with
                                   two qubits removed using spin symmetries.
        WARNING:
                Reorders orbitals from the default even-odd ordering to all
                spin-up orbitals, then all spin-down orbitals.
        Raises:
                ValueError if fermion_hamiltonian isn't of the type
                FermionOperator, or active_orbitals isn't an integer,
                or active_fermions isn't an integer.

        Notes: This function reorders the spin orbitals as all spin-up, then
               all spin-down. It uses the OpenFermion bravyi_kitaev_tree
               mapping, rather than the bravyi-kitaev mapping.
               Caution advised when using with a Fermi-Hubbard Hamiltonian;
               this technique correctly reduces the Hamiltonian only for the
               lowest energy even and odd fermion number states, not states
               with an arbitrary number of fermions.
    """
    # Catch errors if inputs are of wrong type.
    if type(fermion_hamiltonian) is not FermionOperator:
        raise ValueError('Supplied operator should be an instance '
                         'of FermionOperator class')
    if type(active_orbitals) is not int:
        raise ValueError('Number of active orbitals should be an integer.')
    if type(active_fermions) is not int:
        raise ValueError('Number of active fermions should be an integer.')

    # Arrange spins up then down, then BK map to qubit Hamiltonian.
    fermion_hamiltonian_reorder = reorder(fermion_hamiltonian,
                                          up_then_down,
                                          num_modes=active_orbitals)
    qubit_hamiltonian = bravyi_kitaev_tree(fermion_hamiltonian_reorder,
                                           n_qubits=active_orbitals)
    qubit_hamiltonian.compress()

    # Allocates the parity factors for the orbitals as in arXiv:1704.05018.
    remainder = active_fermions % 4
    if remainder == 0:
        parity_final_orb = 1
        parity_middle_orb = 1
    elif remainder == 1:
        parity_final_orb = -1
        parity_middle_orb = -1
    elif remainder == 2:
        parity_final_orb = 1
        parity_middle_orb = -1
    else:
        parity_final_orb = -1
        parity_middle_orb = 1

    # Removes the final qubit, then the middle qubit.
    qubit_hamiltonian = edit_hamiltonian_for_spin(qubit_hamiltonian,
                                                  active_orbitals,
                                                  parity_final_orb)
    qubit_hamiltonian = edit_hamiltonian_for_spin(qubit_hamiltonian,
                                                  active_orbitals / 2,
                                                  parity_middle_orb)
    qubit_hamiltonian = prune_unused_indices(qubit_hamiltonian)

    return qubit_hamiltonian


def edit_hamiltonian_for_spin(qubit_hamiltonian, spin_orbital, orbital_parity):
    """ Removes the Z terms acting on the orbital from the Hamiltonian.
    """
    new_qubit_dict = {}
    for term, coefficient in qubit_hamiltonian.terms.items():
        # If Z acts on the specified orbital, precompute its effect and
        # remove it from the Hamiltonian.
        if (spin_orbital - 1, 'Z') in term:
            new_coefficient = coefficient * orbital_parity
            new_term = tuple(i for i in term if i != (spin_orbital - 1, 'Z'))
            # Make sure to combine terms comprised of the same operators.
            if new_qubit_dict.get(new_term) is None:
                new_qubit_dict[new_term] = new_coefficient
            else:
                old_coefficient = new_qubit_dict.get(new_term)
                new_qubit_dict[new_term] = new_coefficient + old_coefficient
        else:
            # Make sure to combine terms comprised of the same operators.
            if new_qubit_dict.get(term) is None:
                new_qubit_dict[term] = coefficient
            else:
                old_coefficient = new_qubit_dict.get(term)
                new_qubit_dict[term] = coefficient + old_coefficient

    qubit_hamiltonian.terms = new_qubit_dict
    qubit_hamiltonian.compress()

    return qubit_hamiltonian



from itertools import product
import pytest
import numpy

from openfermion.linalg.wedge_product import (wedge,
                                              generate_parity_permutations)


def test_upper_lower_index_size_check():
    n_tensor = numpy.zeros((2, 2, 2))
    m_tensor = numpy.zeros((2, 2))
    with pytest.raises(IndexError):
        wedge(n_tensor, m_tensor, (3, 1), (1, 1))

    with pytest.raises(IndexError):
        wedge(n_tensor, m_tensor, (4, 0), (1, 2))

    with pytest.raises(IndexError):
        wedge(n_tensor, m_tensor, (2, 1), (1, 2))


def test_eye_wedge():
    """
    Test wedge product code through identity wedge

    (p, s)(q, r) - (q, s)(p, r) - (p, r)(q, s) + (q, r)(p, s)
    """
    n_tensor = numpy.eye(2)
    m_tensor = numpy.eye(2)
    true_eye2 = numpy.zeros((2, 2, 2, 2))
    for p, q, r, s in product(range(2), repeat=4):
        true_eye2[p, q, r, s] += (n_tensor[p, s] * m_tensor[q, r] -
                                  n_tensor[q, s] * m_tensor[p, r] -
                                  n_tensor[p, r] * m_tensor[q, s] +
                                  n_tensor[q, r] * m_tensor[p, s])
    true_eye2 /= 2**2
    test_eye2 = wedge(n_tensor, m_tensor, (1, 1), (1, 1))
    assert numpy.allclose(test_eye2, true_eye2)


def test_random_wedge():
    """
    Test wedge product code through identity wedge

    (p, s)(q, r) - (q, s)(p, r) - (p, r)(q, s) + (q, r)(p, s)
    """
    dim = 4
    n_tensor = numpy.random.random((dim, dim))
    m_tensor = numpy.random.random((dim, dim))
    true_eye2 = numpy.zeros(tuple([dim] * 4))
    for p, q, r, s in product(range(dim), repeat=4):
        true_eye2[p, q, r, s] += (n_tensor[p, s] * m_tensor[q, r] -
                                  n_tensor[q, s] * m_tensor[p, r] -
                                  n_tensor[p, r] * m_tensor[q, s] +
                                  n_tensor[q, r] * m_tensor[p, s])
    true_eye2 /= 2**2
    test_eye2 = wedge(n_tensor, m_tensor, (1, 1), (1, 1))
    assert numpy.allclose(test_eye2, true_eye2)


def test_random_two_wedge():
    """
    Test wedge product code with random 4-tensor and a random 2-tensor

    (a, b, c, d) wedge (e, f) -> (a, b, e, f, c, d)

    """
    dim = 3
    n_tensor = numpy.random.random((dim, dim, dim, dim))
    m_tensor = numpy.random.random((dim, dim))
    true_tensor = numpy.zeros(tuple([dim] * 6))
    for a, b, c, d, e, f in product(range(dim), repeat=6):
        for u_perm, u_phase in generate_parity_permutations([a, b, c]):
            for l_perm, l_phase in generate_parity_permutations([f, e, d]):
                true_tensor[a, b, c, d, e, f] += u_phase * l_phase * n_tensor[
                    u_perm[0], u_perm[1], l_perm[1],
                    l_perm[0]] * m_tensor[u_perm[2], l_perm[2]]

    true_tensor /= 6**2
    test_tensor = wedge(n_tensor, m_tensor, (2, 2), (1, 1))
    assert numpy.allclose(test_tensor, true_tensor)


def test_random_two_wedge_two():
    """
    Test wedge product code with a random 4-tensor and a random 4-tensor

    (a, b, h, g) wedge (c, d, f, e) -> (a, b, c, d, e, f, g, h)

    """
    dim = 2
    n_tensor = numpy.random.random((dim, dim, dim, dim))
    m_tensor = numpy.random.random((dim, dim, dim, dim))
    true_tensor = numpy.zeros(tuple([dim] * 8))
    for a, b, c, d, e, f, g, h in product(range(dim), repeat=8):
        for u_perm, u_phase in generate_parity_permutations([a, b, c, d]):
            for l_perm, l_phase in generate_parity_permutations([h, g, f, e]):
                # D_{hgfe}^{abcd} = C [D_{hg}^{ab}D_{fe}^{cd} +
                # -1 * D_{hg}^{ab}D_{ef}^{cd} + ...]
                true_tensor[a, b, c, d, e, f, g, h] += u_phase * l_phase * \
                                                       n_tensor[
                                                           u_perm[0], u_perm[1],
                                                           l_perm[1], l_perm[
                                                               0]] * m_tensor[
                                                           u_perm[2], u_perm[3],
                                                           l_perm[3], l_perm[2]]

    true_tensor /= (4 * 3 * 2)**2
    test_tensor = wedge(n_tensor, m_tensor, (2, 2), (2, 2))
    assert numpy.allclose(test_tensor, true_tensor)


def test_random_1_2_wedge_1_1():
    dim = 3
    n_tensor = numpy.random.random((dim, dim, dim))
    m_tensor = numpy.random.random((dim, dim))
    true_tensor = numpy.zeros(tuple([dim] * 5))
    for a, b, c, d, e in product(range(dim), repeat=5):
        for u_perm, u_phase in generate_parity_permutations([a, b]):
            for l_perm, l_phase in generate_parity_permutations([e, d, c]):
                # D_{edc}^{ab} = C[D_{ed}^{a}D_{c}^{b} - D_{ed}^{b}D_{c}^{a} -
                # D_{ec}^{a}D_{d}^{b} +...]
                true_tensor[a, b, c, d, e] += u_phase * l_phase * n_tensor[
                    u_perm[0], l_perm[2], l_perm[1]] * m_tensor[u_perm[1],
                                                                l_perm[0]]

    true_tensor /= (2 * 3 * 2)
    test_tensor = wedge(n_tensor, m_tensor, (1, 2), (1, 1))
    assert numpy.allclose(test_tensor, true_tensor)


def test_random_2_1_wedge_1_1():
    dim = 3
    n_tensor = numpy.random.random((dim, dim, dim))
    m_tensor = numpy.random.random((dim, dim))
    true_tensor = numpy.zeros(tuple([dim] * 5))
    for a, b, c, d, e in product(range(dim), repeat=5):
        for u_perm, u_phase in generate_parity_permutations([a, b, c]):
            for l_perm, l_phase in generate_parity_permutations([e, d]):
                true_tensor[a, b, c, d, e] += u_phase * l_phase * n_tensor[
                    u_perm[0], u_perm[1], l_perm[0]] * m_tensor[u_perm[2],
                                                                l_perm[1]]

    true_tensor /= (2 * 3 * 2)
    test_tensor = wedge(n_tensor, m_tensor, (2, 1), (1, 1))
    assert numpy.allclose(test_tensor, true_tensor)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import unittest

import numpy

from openfermion.hamiltonians.jellium import (
    dual_basis_jellium_model,
    dual_basis_kinetic,
    dual_basis_potential,
    hypercube_grid_with_given_wigner_seitz_radius_and_filling,
    jellium_model,
    jordan_wigner_dual_basis_jellium,
    plane_wave_kinetic,
    plane_wave_potential,
    wigner_seitz_length_scale,
)
from openfermion.ops.operators import FermionOperator, QubitOperator
from openfermion.transforms.opconversions import jordan_wigner
from openfermion.linalg import get_sparse_operator, eigenspectrum
from openfermion.utils import count_qubits, Grid, is_hermitian


class WignerSeitzRadiusTest(unittest.TestCase):

    def test_wigner_seitz_radius_1d(self):
        wigner_seitz_radius = 3.17
        n_particles = 20
        one_d_test = wigner_seitz_length_scale(wigner_seitz_radius, n_particles,
                                               1)
        self.assertAlmostEqual(one_d_test,
                               n_particles * 2. * wigner_seitz_radius)

    def test_wigner_seitz_radius_2d(self):
        wigner_seitz_radius = 0.5
        n_particles = 3
        two_d_test = wigner_seitz_length_scale(wigner_seitz_radius, n_particles,
                                               2)**2.
        self.assertAlmostEqual(two_d_test,
                               n_particles * numpy.pi * wigner_seitz_radius**2.)

    def test_wigner_seitz_radius_3d(self):
        wigner_seitz_radius = 4.6
        n_particles = 37
        three_d_test = wigner_seitz_length_scale(wigner_seitz_radius,
                                                 n_particles, 3)**3.
        self.assertAlmostEqual(
            three_d_test,
            n_particles * (4. * numpy.pi / 3. * wigner_seitz_radius**3.))

    def test_wigner_seitz_radius_6d(self):
        wigner_seitz_radius = 5.
        n_particles = 42
        six_d_test = wigner_seitz_length_scale(wigner_seitz_radius, n_particles,
                                               6)**6
        self.assertAlmostEqual(
            six_d_test,
            n_particles * (numpy.pi**3 / 6 * wigner_seitz_radius**6))

    def test_wigner_seitz_radius_bad_dimension_not_integer(self):
        with self.assertRaises(ValueError):
            _ = wigner_seitz_length_scale(3, 2, dimension=4.2)

    def test_wigner_seitz_radius_bad_dimension_not_positive(self):
        with self.assertRaises(ValueError):
            _ = wigner_seitz_length_scale(3, 2, dimension=0)


class HypercubeGridTest(unittest.TestCase):

    def test_1d_generation(self):
        dim = 1
        orbitals = 4
        wigner_seitz_radius = 7.

        grid = hypercube_grid_with_given_wigner_seitz_radius_and_filling(
            dim, orbitals, wigner_seitz_radius)
        self.assertEqual(grid.dimensions, 1)
        self.assertEqual(grid.length, (4,))
        self.assertEqual(grid.volume_scale(), orbitals * wigner_seitz_radius)

    def test_generation_away_from_half_filling(self):
        dim = 1
        orbitals = 100
        wigner_seitz_radius = 7.
        filling = 0.2

        grid = hypercube_grid_with_given_wigner_seitz_radius_and_filling(
            dim, orbitals, wigner_seitz_radius, filling_fraction=filling)
        self.assertEqual(grid.dimensions, 1)
        self.assertEqual(grid.length, (100,))
        self.assertAlmostEqual(grid.volume_scale(),
                               orbitals * wigner_seitz_radius / 2.5)

    def test_generation_with_spin(self):
        dim = 2
        orbitals = 4
        wigner_seitz_radius = 10.
        spinless = False

        grid = hypercube_grid_with_given_wigner_seitz_radius_and_filling(
            dim, orbitals, wigner_seitz_radius, spinless=spinless)
        self.assertEqual(grid.dimensions, 2)
        self.assertEqual(grid.length, (4, 4))
        self.assertAlmostEqual(grid.volume_scale(), numpy.pi * 16 * 100.)

    def test_3d_generation_with_rounding(self):
        filling = 0.42
        grid = hypercube_grid_with_given_wigner_seitz_radius_and_filling(
            3, 5, 1., filling_fraction=filling)
        self.assertEqual(grid.dimensions, 3)
        self.assertEqual(grid.length, (5, 5, 5))

        # There are floor(125 * .42) = 52 particles.
        # The volume scale should be 4/3 pi r^3 * the "true" filling fraction.
        self.assertAlmostEqual(grid.volume_scale(),
                               (4. / 3.) * numpy.pi * (5.**3) * (52. / 125))

    def test_raise_ValueError_filling_fraction_too_low(self):
        with self.assertRaises(ValueError):
            _ = hypercube_grid_with_given_wigner_seitz_radius_and_filling(
                3, 5, wigner_seitz_radius=10., filling_fraction=0.005)

    def test_raise_ValueError_filling_fraction_too_high(self):
        with self.assertRaises(ValueError):
            _ = hypercube_grid_with_given_wigner_seitz_radius_and_filling(
                1, 4, wigner_seitz_radius=1., filling_fraction=2.)


class JelliumTest(unittest.TestCase):

    def test_kinetic_integration(self):
        # Compute kinetic energy operator in both momentum and position space.
        grid = Grid(dimensions=2, length=2, scale=3.)
        spinless = False
        momentum_kinetic = plane_wave_kinetic(grid, spinless)
        position_kinetic = dual_basis_kinetic(grid, spinless)

        # Confirm they are Hermitian
        momentum_kinetic_operator = get_sparse_operator(momentum_kinetic)
        self.assertTrue(is_hermitian(momentum_kinetic_operator))

        position_kinetic_operator = get_sparse_operator(position_kinetic)
        self.assertTrue(is_hermitian(position_kinetic_operator))

        # Confirm spectral match and hermiticity
        for length in [2, 3, 4]:
            grid = Grid(dimensions=1, length=length, scale=2.1)
            spinless = False

            momentum_kinetic = plane_wave_kinetic(grid, spinless)
            position_kinetic = dual_basis_kinetic(grid, spinless)

            # Confirm they are Hermitian
            momentum_kinetic_operator = get_sparse_operator(momentum_kinetic)
            self.assertTrue(is_hermitian(momentum_kinetic_operator))

            position_kinetic_operator = get_sparse_operator(position_kinetic)
            self.assertTrue(is_hermitian(position_kinetic_operator))

            # Diagonalize and confirm the same energy.
            jw_momentum = jordan_wigner(momentum_kinetic)
            jw_position = jordan_wigner(position_kinetic)
            momentum_spectrum = eigenspectrum(jw_momentum, 2 * length)
            position_spectrum = eigenspectrum(jw_position, 2 * length)

            # Confirm spectra are the same.
            difference = numpy.amax(
                numpy.absolute(momentum_spectrum - position_spectrum))
            self.assertAlmostEqual(difference, 0.)

    def test_potential_integration(self):
        # Compute potential energy operator in momentum and position space.
        for length in [2, 3]:
            grid = Grid(dimensions=2, length=length, scale=2.)
            spinless = True
            momentum_potential = plane_wave_potential(grid, spinless)
            position_potential = dual_basis_potential(grid, spinless)

            # Confirm they are Hermitian
            momentum_potential_operator = (
                get_sparse_operator(momentum_potential))
            self.assertTrue(is_hermitian(momentum_potential_operator))

            position_potential_operator = (
                get_sparse_operator(position_potential))
            self.assertTrue(is_hermitian(position_potential_operator))

            # Diagonalize and confirm the same energy.
            jw_momentum = jordan_wigner(momentum_potential)
            jw_position = jordan_wigner(position_potential)
            momentum_spectrum = eigenspectrum(jw_momentum)
            position_spectrum = eigenspectrum(jw_position)

            # Confirm spectra are the same.
            difference = numpy.amax(
                numpy.absolute(momentum_spectrum - position_spectrum))
            self.assertAlmostEqual(difference, 0.)

    def test_model_integration(self):
        # Compute Hamiltonian in both momentum and position space.
        for length in [2, 3]:
            grid = Grid(dimensions=2, length=length, scale=1.0)
            spinless = True
            momentum_hamiltonian = jellium_model(grid, spinless, True)
            position_hamiltonian = jellium_model(grid, spinless, False)

            # Confirm they are Hermitian
            momentum_hamiltonian_operator = (
                get_sparse_operator(momentum_hamiltonian))
            self.assertTrue(is_hermitian(momentum_hamiltonian_operator))

            position_hamiltonian_operator = (
                get_sparse_operator(position_hamiltonian))
            self.assertTrue(is_hermitian(position_hamiltonian_operator))

            # Diagonalize and confirm the same energy.
            jw_momentum = jordan_wigner(momentum_hamiltonian)
            jw_position = jordan_wigner(position_hamiltonian)
            momentum_spectrum = eigenspectrum(jw_momentum)
            position_spectrum = eigenspectrum(jw_position)

            # Confirm spectra are the same.
            difference = numpy.amax(
                numpy.absolute(momentum_spectrum - position_spectrum))
            self.assertAlmostEqual(difference, 0.)

    def test_model_integration_with_constant(self):
        # Compute Hamiltonian in both momentum and position space.
        length_scale = 0.7
        for length in [2, 3]:
            grid = Grid(dimensions=2, length=length, scale=length_scale)
            spinless = True

            # Include Madelung constant in the momentum but not the position
            # Hamiltonian.
            momentum_hamiltonian = jellium_model(grid,
                                                 spinless,
                                                 True,
                                                 include_constant=True)
            position_hamiltonian = jellium_model(grid, spinless, False)

            # Confirm they are Hermitian
            momentum_hamiltonian_operator = (
                get_sparse_operator(momentum_hamiltonian))
            self.assertTrue(is_hermitian(momentum_hamiltonian_operator))

            position_hamiltonian_operator = (
                get_sparse_operator(position_hamiltonian))
            self.assertTrue(is_hermitian(position_hamiltonian_operator))

            # Diagonalize and confirm the same energy.
            jw_momentum = jordan_wigner(momentum_hamiltonian)
            jw_position = jordan_wigner(position_hamiltonian)
            momentum_spectrum = eigenspectrum(jw_momentum)
            position_spectrum = eigenspectrum(jw_position)

            # Confirm momentum spectrum is shifted 2.8372/length_scale higher.
            max_difference = numpy.amax(momentum_spectrum - position_spectrum)
            min_difference = numpy.amax(momentum_spectrum - position_spectrum)
            self.assertAlmostEqual(max_difference, 2.8372 / length_scale)
            self.assertAlmostEqual(min_difference, 2.8372 / length_scale)

    def test_coefficients(self):
        # Test that the coefficients post-JW transform are as claimed in paper.
        grid = Grid(dimensions=2, length=3, scale=2.)
        spinless = 1
        n_orbitals = grid.num_points
        n_qubits = (2**(1 - spinless)) * n_orbitals
        volume = grid.volume_scale()

        # Kinetic operator.
        kinetic = dual_basis_kinetic(grid, spinless)
        qubit_kinetic = jordan_wigner(kinetic)

        # Potential operator.
        potential = dual_basis_potential(grid, spinless)
        qubit_potential = jordan_wigner(potential)

        # Check identity.
        identity = tuple()
        kinetic_coefficient = qubit_kinetic.terms[identity]
        potential_coefficient = qubit_potential.terms[identity]

        paper_kinetic_coefficient = 0.
        paper_potential_coefficient = 0.
        for indices in grid.all_points_indices():
            momenta = grid.momentum_vector(indices)
            paper_kinetic_coefficient += float(n_qubits) * momenta.dot(
                momenta) / float(4. * n_orbitals)

            if momenta.any():
                potential_contribution = -numpy.pi * float(n_qubits) / float(
                    2. * momenta.dot(momenta) * volume)
                paper_potential_coefficient += potential_contribution

        self.assertAlmostEqual(kinetic_coefficient, paper_kinetic_coefficient)
        self.assertAlmostEqual(potential_coefficient,
                               paper_potential_coefficient)

        # Check Zp.
        for p in range(n_qubits):
            zp = ((p, 'Z'),)
            kinetic_coefficient = qubit_kinetic.terms[zp]
            potential_coefficient = qubit_potential.terms[zp]

            paper_kinetic_coefficient = 0.
            paper_potential_coefficient = 0.
            for indices in grid.all_points_indices():
                momenta = grid.momentum_vector(indices)
                paper_kinetic_coefficient -= momenta.dot(momenta) / float(
                    4. * n_orbitals)

                if momenta.any():
                    potential_contribution = numpy.pi / float(
                        momenta.dot(momenta) * volume)
                    paper_potential_coefficient += potential_contribution

            self.assertAlmostEqual(kinetic_coefficient,
                                   paper_kinetic_coefficient)
            self.assertAlmostEqual(potential_coefficient,
                                   paper_potential_coefficient)

        # Check Zp Zq.
        if spinless:
            spins = [None]

        for indices_a in grid.all_points_indices():
            for indices_b in grid.all_points_indices():

                potential_coefficient = 0.
                paper_kinetic_coefficient = 0.
                paper_potential_coefficient = 0.

                position_a = grid.position_vector(indices_a)
                position_b = grid.position_vector(indices_b)
                differences = position_b - position_a

                for spin_a in spins:
                    for spin_b in spins:

                        p = grid.orbital_id(indices_a, spin_a)
                        q = grid.orbital_id(indices_b, spin_b)

                        if p == q:
                            continue

                        zpzq = ((min(p, q), 'Z'), (max(p, q), 'Z'))
                        if zpzq in qubit_potential.terms:
                            potential_coefficient = qubit_potential.terms[zpzq]

                        for indices_c in grid.all_points_indices():
                            momenta = grid.momentum_vector(indices_c)

                            if momenta.any():
                                potential_contribution = numpy.pi * numpy.cos(
                                    differences.dot(momenta)) / float(
                                        momenta.dot(momenta) * volume)
                                paper_potential_coefficient += (
                                    potential_contribution)

                        self.assertAlmostEqual(potential_coefficient,
                                               paper_potential_coefficient)

    def test_jordan_wigner_dual_basis_jellium(self):
        # Parameters.
        grid = Grid(dimensions=2, length=3, scale=1.)
        spinless = True

        # Compute fermionic Hamiltonian. Include then subtract constant.
        fermion_hamiltonian = dual_basis_jellium_model(grid,
                                                       spinless,
                                                       include_constant=True)
        qubit_hamiltonian = jordan_wigner(fermion_hamiltonian)
        qubit_hamiltonian -= QubitOperator((), 2.8372)

        # Compute Jordan-Wigner Hamiltonian.
        test_hamiltonian = jordan_wigner_dual_basis_jellium(grid, spinless)

        # Make sure Hamiltonians are the same.
        self.assertTrue(test_hamiltonian == qubit_hamiltonian)

        # Check number of terms.
        n_qubits = count_qubits(qubit_hamiltonian)
        if spinless:
            paper_n_terms = 1 - .5 * n_qubits + 1.5 * (n_qubits**2)

        num_nonzeros = sum(
            1 for coeff in qubit_hamiltonian.terms.values() if coeff != 0.0)
        self.assertTrue(num_nonzeros <= paper_n_terms)

    def test_jordan_wigner_dual_basis_jellium_constant_shift(self):
        length_scale = 0.6
        grid = Grid(dimensions=2, length=3, scale=length_scale)
        spinless = True

        hamiltonian_without_constant = jordan_wigner_dual_basis_jellium(
            grid, spinless, include_constant=False)
        hamiltonian_with_constant = jordan_wigner_dual_basis_jellium(
            grid, spinless, include_constant=True)

        difference = hamiltonian_with_constant - hamiltonian_without_constant
        expected = QubitOperator('') * (2.8372 / length_scale)

        self.assertTrue(expected == difference)

    def test_plane_wave_energy_cutoff(self):
        grid = Grid(dimensions=1, length=5, scale=1.0)
        spinless = True
        e_cutoff = 20.0

        hamiltonian_1 = jellium_model(grid, spinless, True, False)
        jw_1 = jordan_wigner(hamiltonian_1)
        spectrum_1 = eigenspectrum(jw_1)

        hamiltonian_2 = jellium_model(grid, spinless, True, False, e_cutoff)
        jw_2 = jordan_wigner(hamiltonian_2)
        spectrum_2 = eigenspectrum(jw_2)

        max_diff = numpy.amax(numpy.absolute(spectrum_1 - spectrum_2))
        self.assertGreater(max_diff, 0.)

    def test_plane_wave_period_cutoff(self):
        # TODO: After figuring out the correct formula for period cutoff for
        #     dual basis, change period_cutoff to default, and change
        #     hamiltonian_1 to a real jellium_model for real integration test.

        grid = Grid(dimensions=2, length=2, scale=1.0)
        spinless = True
        period_cutoff = 0.

        hamiltonian_1 = FermionOperator()
        jw_1 = jordan_wigner(hamiltonian_1)
        spectrum_1 = eigenspectrum(jw_1)

        hamiltonian_2 = jellium_model(grid, spinless, True, False, None, True,
                                      period_cutoff)
        jw_2 = jordan_wigner(hamiltonian_2)
        spectrum_2 = eigenspectrum(jw_2)

        max_diff = numpy.amax(numpy.absolute(spectrum_1 - spectrum_2))
        self.assertGreater(max_diff, 0.)

        # TODO: This is only for code coverage. Remove after having real
        #     integration test.
        jellium_model(grid, spinless, True, False, None, True)
        jellium_model(grid, spinless, False, False, None, True)



import pytest
import numpy as np
from openfermion.third_party.representability._dualbasis import \
    DualBasisElement, DualBasis


def test_dualbasis_element_init():
    dbe = DualBasisElement()
    assert isinstance(dbe, DualBasisElement)

    dbe = DualBasisElement(tensor_names=['A'] * 5,
                           tensor_coeffs=[1] * 5,
                           tensor_elements=[(i, i) for i in range(5)])

    assert dbe.primal_tensors_names == ['A'] * 5
    assert dbe.primal_coeffs == [1] * 5
    assert dbe.primal_elements == [(i, i) for i in range(5)]
    assert dbe.constant_bias == 0
    assert dbe.dual_scalar == 0


def test_daulbasis_init():
    db = DualBasis()
    assert isinstance(db, DualBasis)
    dbe = DualBasisElement(tensor_names=['A'] * 5,
                           tensor_coeffs=[1] * 5,
                           tensor_elements=[(i, i) for i in range(5)])
    db = DualBasis(elements=[dbe])
    assert db[0] == dbe
    assert len(db) == 1
    for tdbe in db:
        assert tdbe == dbe

    db2 = db + dbe
    assert isinstance(db2, DualBasis)
    assert len(db2) == 2

    tdbe = DualBasisElement(tensor_names=['B'] * 5,
                            tensor_coeffs=[1] * 5,
                            tensor_elements=[(i, i) for i in range(5)])
    tdb = DualBasis(elements=[tdbe])
    db3 = db + tdb
    assert isinstance(db3, DualBasis)
    assert len(db3) == 2

    db4 = tdbe + db
    assert len(db4) == 2

    with pytest.raises(TypeError):
        _ = DualBasis(elements=[tdbe, 4])

    with pytest.raises(TypeError):
        db = DualBasis(elements=[tdbe])
        _ = db + 4


def test_dbe_element_add():
    de = DualBasisElement()
    de_2 = DualBasisElement()
    db_0 = de + de_2
    assert isinstance(db_0, DualBasis)
    db_0 = de_2 + de
    assert isinstance(db_0, DualBasis)
    db_1 = db_0 + db_0
    assert isinstance(db_1, DualBasis)

    with pytest.raises(TypeError):
        _ = de + 2


def test_dbe_string():
    dbe = DualBasisElement(tensor_names=['A'] * 5,
                           tensor_coeffs=[1] * 5,
                           tensor_elements=[(i, i) for i in range(5)])

    assert dbe.id() == "A(0,0)	A(1,1)	A(2,2)	A(3,3)	A(4,4)\t"


def test_dbe_iterator():
    dbe = DualBasisElement(tensor_names=['A'] * 5,
                           tensor_coeffs=[1] * 5,
                           tensor_elements=[(i, i) for i in range(5)])
    for idx, (t, v, c) in enumerate(dbe):
        assert t == 'A'
        assert v == (idx, idx)
        assert c == 1


def test_add_element():
    with pytest.raises(TypeError):
        dbe = DualBasisElement()
        dbe.add_element('cckk', [1, 2, 3, 4], 0.5)

    with pytest.raises(TypeError):
        dbe = DualBasisElement()
        dbe.add_element(5, (1, 2, 3, 4), 0.5)

    with pytest.raises(TypeError):
        dbe = DualBasisElement()
        dbe.add_element('a', (1, 2, 3, 4), 'a')

    dbe = DualBasisElement()
    dbe.add_element('cckk', (1, 2, 3, 4), 0.5)
    dbe.constant_bias = 0.33

    assert dbe.primal_coeffs == [0.5]
    assert dbe.primal_tensors_names == ['cckk']
    assert dbe.primal_elements == [(1, 2, 3, 4)]

    dbe.add_element('ck', (0, 1), 1)
    assert dbe.primal_coeffs == [0.5, 1]
    assert dbe.primal_tensors_names == ['cckk', 'ck']
    assert dbe.primal_elements == [(1, 2, 3, 4), (0, 1)]

    dbe2 = DualBasisElement()
    with pytest.raises(TypeError):
        dbe2.join_elements((1, 1))

    dbe2.constant_bias = 0.25
    dbe2.add_element('cckk', (1, 2, 3, 4), 0.5)
    dbe3 = dbe2.join_elements(dbe)
    assert np.isclose(dbe3.constant_bias, 0.58)

    assert set(dbe3.primal_elements) == {(0, 1), (1, 2, 3, 4)}
    assert np.allclose(dbe3.primal_coeffs, [1, 1])
    assert set(dbe3.primal_tensors_names) == {'ck', 'cckk'}


def test_simplify():
    i, j, k, l = 0, 1, 2, 3
    names = ['opdm'] * 3 + ['oqdm']
    elements = [(i, j), (i, j), (i, l), (l, k)]
    coeffs = [1, 1, 0.25, 0.3]
    dbe = DualBasisElement(tensor_names=names,
                           tensor_elements=elements,
                           tensor_coeffs=coeffs)
    dbe.simplify()
    assert len(dbe.primal_tensors_names) == 3
    assert set(dbe.primal_coeffs) == {2, 0.25, 0.3}
    assert set(dbe.primal_tensors_names) == {'opdm', 'oqdm'}
    assert set(dbe.primal_elements) == {(0, 1), (0, 3), (3, 2)}



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""This module provides tools for simulating rank deficient operators."""

import itertools
import numpy

from openfermion.config import EQ_TOLERANCE
import openfermion.utils as utils


def get_chemist_two_body_coefficients(two_body_coefficients, spin_basis=True):
    r"""Convert two-body operator coefficients to low rank tensor.

    The input is a two-body fermionic Hamiltonian expressed as
    :math:`\sum_{pqrs} h_{pqrs} a^\dagger_p a^\dagger_q a_r a_s`

    We will convert this to the chemistry convention expressing it as
    :math:`\sum_{pqrs} g_{pqrs} a^\dagger_p a_q a^\dagger_r a_s`
    but without the spin degree of freedom.

    In the process of performing this conversion, constants and one-body
    terms come out, which will be returned as well.

    Args:
        two_body_coefficients (ndarray): an N x N x N x N
            numpy array giving the :math:`h_{pqrs}` tensor.
        spin_basis (bool): True if the two-body terms are passed in spin
            orbital basis. False if already in spatial orbital basis.

    Returns:
        one_body_correction (ndarray): an N x N array of floats giving
            coefficients of the :math:`a^\dagger_p a_q` terms that come out.
        chemist_two_body_coefficients (ndarray): an N x N x N x N numpy array
            giving the :math:`g_{pqrs}` tensor in chemist notation.

    Raises:
        TypeError: Input must be two-body number conserving
            FermionOperator or InteractionOperator.
    """
    # Initialize.
    n_orbitals = two_body_coefficients.shape[0]
    chemist_two_body_coefficients = numpy.transpose(two_body_coefficients,
                                                    [0, 3, 1, 2])

    # If the specification was in spin-orbitals, chop down to spatial orbitals
    # assuming a spin-symmetric interaction.
    if spin_basis:
        n_orbitals = n_orbitals // 2
        alpha_indices = list(range(0, n_orbitals * 2, 2))
        beta_indices = list(range(1, n_orbitals * 2, 2))
        chemist_two_body_coefficients = chemist_two_body_coefficients[numpy.ix_(
            alpha_indices, alpha_indices, beta_indices, beta_indices)]

    # Determine a one body correction in the spin basis from spatial basis.
    one_body_correction = numpy.zeros((2 * n_orbitals, 2 * n_orbitals), complex)
    for p, q, r, s in itertools.product(range(n_orbitals), repeat=4):
        for sigma, tau in itertools.product(range(2), repeat=2):
            if (q == r) and (sigma == tau):
                one_body_correction[2 * p + sigma, 2 * s + tau] -= (
                    chemist_two_body_coefficients[p, q, r, s])

    # Return.
    return one_body_correction, chemist_two_body_coefficients


def low_rank_two_body_decomposition(two_body_coefficients,
                                    truncation_threshold=1e-8,
                                    final_rank=None,
                                    spin_basis=True):
    r"""Convert two-body operator into sum of squared one-body operators.

    As in arXiv:1808.02625, this function decomposes
    :math:`\sum_{pqrs} h_{pqrs} a^\dagger_p a^\dagger_q a_r a_s` as
    :math:`\sum_{l} \lambda_l (\sum_{pq} g_{lpq} a^\dagger_p a_q)^2`
    l is truncated to take max value L so that
    :math:`\sum_{l=0}^{L-1} (\sum_{pq} |g_{lpq}|)^2 |\lambda_l| < x`

    Args:
        two_body_coefficients (ndarray): an N x N x N x N
            numpy array giving the :math:`h_{pqrs}` tensor.
            This tensor must be 8-fold symmetric (real integrals).
        truncation_threshold (optional Float): the value of x, above.
        final_rank (optional int): if provided, this specifies the value of
            L at which to truncate. This overrides truncation_threshold.
        spin_basis (bool): True if the two-body terms are passed in spin
            orbital basis.  False if already in spatial orbital basis.

    Returns:
        eigenvalues (ndarray of floats): length L array
            giving the :math:`\lambda_l`.
        one_body_squares (ndarray of floats): L x N x N array of floats
            corresponding to the value of :math:`g_{pql}`.
        one_body_correction (ndarray): One-body correction terms that result
            from reordering to chemist ordering, in spin-orbital basis.
        truncation_value (float): after truncation, this is the value
            :math:`\sum_{l=0}^{L-1} (\sum_{pq} |g_{lpq}|)^2 |\lambda_l| < x`

    Raises:
        TypeError: Invalid two-body coefficient tensor specification.
    """
    # Initialize N^2 by N^2 interaction array.
    one_body_correction, chemist_two_body_coefficients = (
        get_chemist_two_body_coefficients(two_body_coefficients, spin_basis))
    n_orbitals = chemist_two_body_coefficients.shape[0]
    full_rank = n_orbitals**2
    interaction_array = numpy.reshape(chemist_two_body_coefficients,
                                      (full_rank, full_rank))

    # Make sure interaction array is symmetric and real.
    asymmetry = numpy.sum(
        numpy.absolute(interaction_array - interaction_array.transpose()))
    imaginary_norm = numpy.sum(numpy.absolute(interaction_array.imag))
    if asymmetry > EQ_TOLERANCE or imaginary_norm > EQ_TOLERANCE:
        raise TypeError('Invalid two-body coefficient tensor specification.')

    # Decompose with exact diagonalization.
    eigenvalues, eigenvectors = numpy.linalg.eigh(interaction_array)

    # Get one-body squares and compute weights.
    term_weights = numpy.zeros(full_rank)
    one_body_squares = numpy.zeros((full_rank, 2 * n_orbitals, 2 * n_orbitals),
                                   complex)

    # Reshape and add spin back in.
    for l in range(full_rank):
        one_body_squares[l] = numpy.kron(
            numpy.reshape(eigenvectors[:, l], (n_orbitals, n_orbitals)),
            numpy.eye(2))
        term_weights[l] = abs(eigenvalues[l]) * numpy.sum(
            numpy.absolute(one_body_squares[l]))**2

    # Sort by weight.
    indices = numpy.argsort(term_weights)[::-1]
    eigenvalues = eigenvalues[indices]
    term_weights = term_weights[indices]
    one_body_squares = one_body_squares[indices]

    # Determine upper-bound on truncation errors that would occur.
    cumulative_error_sum = numpy.cumsum(term_weights)
    truncation_errors = cumulative_error_sum[-1] - cumulative_error_sum

    # Optionally truncate rank and return.
    if final_rank is None:
        max_rank = 1 + numpy.argmax(truncation_errors <= truncation_threshold)
    else:
        max_rank = final_rank
    truncation_value = truncation_errors[max_rank - 1]
    return (eigenvalues[:max_rank], one_body_squares[:max_rank],
            one_body_correction, truncation_value)


def prepare_one_body_squared_evolution(one_body_matrix, spin_basis=True):
    r"""Get Givens angles and DiagonalHamiltonian to simulate squared one-body.

    The goal here will be to prepare to simulate evolution under
    :math:`(\sum_{pq} h_{pq} a^\dagger_p a_q)^2` by decomposing as
    :math:`R e^{-i \sum_{pq} V_{pq} n_p n_q} R^\dagger' where
    :math:`R` is a basis transformation matrix.

    TODO: Add option for truncation based on one-body eigenvalues.

    Args:
        one_body_matrix (ndarray of floats): an N by N array storing the
            coefficients of a one-body operator to be squared. For instance,
            in the above the elements of this matrix are :math:`h_{pq}`.
        spin_basis (bool): Whether the matrix is passed in the
            spin orbital basis.

    Returns:
        density_density_matrix(ndarray of floats) an N by N array storing
            the diagonal two-body coefficeints :math:`V_{pq}` above.
        basis_transformation_matrix (ndarray of floats) an N by N array
            storing the values of the basis transformation.

    Raises:
        ValueError: one_body_matrix is not Hermitian.
    """
    # If the specification was in spin-orbitals, chop back down to spatial orbs
    # assuming a spin-symmetric interaction
    if spin_basis:
        n_modes = one_body_matrix.shape[0]
        alpha_indices = list(range(0, n_modes, 2))
        one_body_matrix = one_body_matrix[numpy.ix_(alpha_indices,
                                                    alpha_indices)]

    # Diagonalize the one-body matrix.
    if utils.is_hermitian(one_body_matrix):
        eigenvalues, eigenvectors = numpy.linalg.eigh(one_body_matrix)
    else:
        raise ValueError('one_body_matrix is not Hermitian.')
    basis_transformation_matrix = numpy.conjugate(eigenvectors.transpose())

    # If the specification was in spin-orbitals, expand back
    if spin_basis:
        basis_transformation_matrix = numpy.kron(basis_transformation_matrix,
                                                 numpy.eye(2))
        eigenvalues = numpy.kron(eigenvalues, numpy.ones(2))

    # Obtain the diagonal two-body matrix.
    density_density_matrix = numpy.outer(eigenvalues, eigenvalues)

    # Return.
    return density_density_matrix, basis_transformation_matrix



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from .binary_polynomial import (
    BinaryPolynomial,
    BinaryPolynomialError,
    binary_sum_rule,
)

from .boson_operator import BosonOperator

from .fermion_operator import FermionOperator

from .ising_operator import IsingOperator

from .majorana_operator import MajoranaOperator

from .quad_operator import QuadOperator

from .qubit_operator import QubitOperator

from .symbolic_operator import SymbolicOperator

# out of alphabetical order to avoid circular import
from .binary_code import (
    double_decoding,
    shift_decoder,
    BinaryCode,
    BinaryCodeError,
)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Fourier transforms on operators"""

import numpy

from openfermion.ops.operators import FermionOperator


def _fourier_transform_helper(hamiltonian, grid, spinless, phase_factor,
                              vec_func_1, vec_func_2):
    hamiltonian_t = FermionOperator.zero()
    normalize_factor = numpy.sqrt(1.0 / float(grid.num_points))

    for term in hamiltonian.terms:
        transformed_term = FermionOperator.identity()
        for ladder_op_mode, ladder_op_type in term:
            indices_1 = grid.grid_indices(ladder_op_mode, spinless)
            vec1 = vec_func_1(indices_1)
            new_basis = FermionOperator.zero()
            for indices_2 in grid.all_points_indices():
                vec2 = vec_func_2(indices_2)
                spin = None if spinless else ladder_op_mode % 2
                orbital = grid.orbital_id(indices_2, spin)
                exp_index = phase_factor * 1.0j * numpy.dot(vec1, vec2)
                if ladder_op_type == 1:
                    exp_index *= -1.0

                element = FermionOperator(((orbital, ladder_op_type),),
                                          numpy.exp(exp_index))
                new_basis += element

            new_basis *= normalize_factor
            transformed_term *= new_basis

        # Coefficient.
        transformed_term *= hamiltonian.terms[term]

        hamiltonian_t += transformed_term

    return hamiltonian_t


def fourier_transform(hamiltonian, grid, spinless):
    r"""Apply Fourier transform to change hamiltonian in plane wave basis.

    .. math::

        c^\dagger_v = \sqrt{1/N} \sum_m {a^\dagger_m \exp(-i k_v r_m)}
        c_v = \sqrt{1/N} \sum_m {a_m \exp(i k_v r_m)}

    Args:
        hamiltonian (FermionOperator): The hamiltonian in plane wave basis.
        grid (Grid): The discretization to use.
        spinless (bool): Whether to use the spinless model or not.

    Returns:
        FermionOperator: The fourier-transformed hamiltonian.
    """
    return _fourier_transform_helper(hamiltonian=hamiltonian,
                                     grid=grid,
                                     spinless=spinless,
                                     phase_factor=+1,
                                     vec_func_1=grid.momentum_vector,
                                     vec_func_2=grid.position_vector)


def inverse_fourier_transform(hamiltonian, grid, spinless):
    r"""Apply inverse Fourier transform to change hamiltonian in
    plane wave dual basis.

    .. math::

        a^\dagger_v = \sqrt{1/N} \sum_m {c^\dagger_m \exp(i k_v r_m)}
        a_v = \sqrt{1/N} \sum_m {c_m \exp(-i k_v r_m)}

    Args:
        hamiltonian (FermionOperator):
            The hamiltonian in plane wave dual basis.
        grid (Grid): The discretization to use.
        spinless (bool): Whether to use the spinless model or not.

    Returns:
        FermionOperator: The inverse-fourier-transformed hamiltonian.
    """
    return _fourier_transform_helper(hamiltonian=hamiltonian,
                                     grid=grid,
                                     spinless=spinless,
                                     phase_factor=-1,
                                     vec_func_1=grid.position_vector,
                                     vec_func_2=grid.momentum_vector)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from .davidson import (
    Davidson,
    DavidsonOptions,
    DavidsonError,
    QubitDavidson,
    SparseDavidson,
    generate_random_vectors,
    append_random_vectors,
    orthonormalize,
)

from .erpa import (
    erpa_eom_hamiltonian,
    singlet_erpa,
)

from .givens_rotations import (
    givens_decomposition,
    givens_rotate,
    double_givens_rotate,
    givens_decomposition_square,
    givens_matrix_elements,
    fermionic_gaussian_decomposition,
    swap_rows,
    swap_columns,
)

from .linear_qubit_operator import (
    generate_linear_qubit_operator,
    LinearQubitOperator,
    LinearQubitOperatorOptions,
    ParallelLinearQubitOperator,
)

from .rdm_reconstruction import (
    valdemoro_reconstruction,)

from .sparse_tools import (
    wrapped_kronecker,
    kronecker_operators,
    jordan_wigner_ladder_sparse,
    jordan_wigner_sparse,
    qubit_operator_sparse,
    eigenspectrum,
    get_linear_qubit_operator_diagonal,
    jw_configuration_state,
    jw_hartree_fock_state,
    jw_number_indices,
    jw_sz_indices,
    jw_number_restrict_operator,
    jw_sz_restrict_operator,
    jw_number_restrict_state,
    jw_sz_restrict_state,
    jw_get_ground_state_at_particle_number,
    jw_sparse_givens_rotation,
    jw_sparse_particle_hole_transformation_last_mode,
    get_density_matrix,
    get_ground_state,
    sparse_eigenspectrum,
    expectation,
    variance,
    expectation_computational_basis_state,
    expectation_db_operator_with_pw_basis_state,
    expectation_one_body_db_operator_computational_basis_state,
    expectation_two_body_db_operator_computational_basis_state,
    expectation_three_body_db_operator_computational_basis_state,
    get_gap,
    inner_product,
    boson_ladder_sparse,
    single_quad_op_sparse,
    boson_operator_sparse,
    get_sparse_operator,
    get_number_preserving_sparse_operator,
)

from .wave_fitting import (
    fit_known_frequencies,
    prony,
)

from .wedge_product import (
    generate_parity_permutations,
    wedge,
)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Optional, Sequence, Type, Union

import sympy

import cirq

_setup_code = ('import cirq\n'
               'import numpy as np\n'
               'import sympy\n'
               'import openfermion\n')


def assert_equivalent_repr(value: Any, *,
                           setup_code: str = _setup_code) -> None:
    """Checks that eval(repr(v)) == v.

    Args:
        value: A value whose repr should be evaluatable python
            code that produces an equivalent value.
        setup_code: Code that must be executed before the repr can be evaluated.
            Ideally this should just be a series of 'import' lines.
    """
    cirq.testing.assert_equivalent_repr(value, setup_code=setup_code)


def assert_implements_consistent_protocols(
        val: Any,
        *,
        exponents: Sequence[Any] = (0, 1, -1, 0.5, 0.25, -0.5, 0.1,
                                    sympy.Symbol('s')),
        qubit_count: Optional[int] = None,
        ignoring_global_phase: bool = False,
        setup_code: str = _setup_code,
        global_vals: Optional[Dict[str, Any]] = None,
        local_vals: Optional[Dict[str, Any]] = None) -> None:
    """Checks that a value is internally consistent and has a good __repr__."""

    cirq.testing.assert_implements_consistent_protocols(
        val,
        exponents=exponents,
        qubit_count=qubit_count,
        ignoring_global_phase=ignoring_global_phase,
        setup_code=setup_code,
        global_vals=global_vals,
        local_vals=local_vals)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Optional, Tuple

import numpy
import pytest
import scipy.sparse.linalg
import cirq

import openfermion
from openfermion import simulate_trotter
from openfermion.circuits.trotter import (
    LINEAR_SWAP_NETWORK,
    LOW_RANK,
    LowRankTrotterAlgorithm,
    SPLIT_OPERATOR,
    TrotterAlgorithm,
)
from openfermion.circuits.trotter.trotter_algorithm import Hamiltonian


def fidelity(state1, state2):
    return abs(numpy.dot(state1, numpy.conjugate(state2)))**2


def produce_simulation_test_parameters(hamiltonian: Hamiltonian,
                                       time: float,
                                       seed: Optional[int] = None
                                      ) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """Produce objects for testing Hamiltonian simulation.

    Produces a random initial state and evolves it under the given Hamiltonian
    for the specified amount of time. Returns the initial state and final
    state.

    Args:
        hamiltonian: The Hamiltonian to evolve under.
        time: The time to evolve for
        seed: An RNG seed.
    """

    n_qubits = openfermion.count_qubits(hamiltonian)

    # Construct a random initial state
    initial_state = openfermion.haar_random_vector(2**n_qubits, seed)

    # Simulate exact evolution
    hamiltonian_sparse = openfermion.get_sparse_operator(hamiltonian)
    exact_state = scipy.sparse.linalg.expm_multiply(
        -1j * time * hamiltonian_sparse, initial_state)

    # Make sure the time is not too small
    assert fidelity(exact_state, initial_state) < .95

    return initial_state, exact_state


# Produce test parameters

longer_time = 1.0
long_time = 0.1
short_time = 0.05

# 5-qubit random DiagonalCoulombHamiltonian
diag_coul_hamiltonian = openfermion.random_diagonal_coulomb_hamiltonian(
    5, real=False, seed=65233)
diag_coul_initial_state, diag_coul_exact_state = (
    produce_simulation_test_parameters(diag_coul_hamiltonian,
                                       long_time,
                                       seed=49075))

# Hubbard model, reordered
hubbard_model = openfermion.fermi_hubbard(2, 2, 1.0, 4.0)
hubbard_model = openfermion.reorder(hubbard_model, openfermion.up_then_down)
hubbard_hamiltonian = openfermion.get_diagonal_coulomb_hamiltonian(
    hubbard_model)
hubbard_initial_state, hubbard_exact_state = (
    produce_simulation_test_parameters(hubbard_hamiltonian,
                                       long_time,
                                       seed=8200))

# 4-qubit H2 2-2 with bond length 0.7414
bond_length = 0.7414
geometry = [('H', (0., 0., 0.)), ('H', (0., 0., bond_length))]
h2_hamiltonian = openfermion.load_molecular_hamiltonian(geometry, 'sto-3g', 1,
                                                        format(bond_length), 2,
                                                        2)
h2_initial_state, h2_exact_state = produce_simulation_test_parameters(
    h2_hamiltonian, longer_time, seed=44303)

# 4-qubit LiH 2-2 with bond length 1.45
bond_length = 1.45
geometry = [('Li', (0., 0., 0.)), ('H', (0., 0., bond_length))]
lih_hamiltonian = openfermion.load_molecular_hamiltonian(
    geometry, 'sto-3g', 1, format(bond_length), 2, 2)
lih_initial_state, lih_exact_state = produce_simulation_test_parameters(
    lih_hamiltonian, longer_time, seed=54458)


@pytest.mark.parametrize(
    'hamiltonian, time, initial_state, exact_state, order, n_steps, '
    'algorithm, result_fidelity', [
        (diag_coul_hamiltonian, long_time, diag_coul_initial_state,
         diag_coul_exact_state, 0, 5, None, .99),
        (diag_coul_hamiltonian, long_time, diag_coul_initial_state,
         diag_coul_exact_state, 0, 12, None, .999),
        (diag_coul_hamiltonian, long_time, diag_coul_initial_state,
         diag_coul_exact_state, 1, 1, LINEAR_SWAP_NETWORK, .99),
        (diag_coul_hamiltonian, long_time, diag_coul_initial_state,
         diag_coul_exact_state, 2, 1, LINEAR_SWAP_NETWORK, .99999),
        (diag_coul_hamiltonian, long_time, diag_coul_initial_state,
         diag_coul_exact_state, 0, 3, SPLIT_OPERATOR, .99),
        (diag_coul_hamiltonian, long_time, diag_coul_initial_state,
         diag_coul_exact_state, 0, 6, SPLIT_OPERATOR, .999),
        (diag_coul_hamiltonian, long_time, diag_coul_initial_state,
         diag_coul_exact_state, 1, 1, SPLIT_OPERATOR, .99),
        (diag_coul_hamiltonian, long_time, diag_coul_initial_state,
         diag_coul_exact_state, 2, 1, SPLIT_OPERATOR, .99999),
        (hubbard_hamiltonian, long_time, hubbard_initial_state,
         hubbard_exact_state, 0, 3, SPLIT_OPERATOR, .999),
        (hubbard_hamiltonian, long_time, hubbard_initial_state,
         hubbard_exact_state, 0, 6, SPLIT_OPERATOR, .9999),
        (h2_hamiltonian, longer_time, h2_initial_state, h2_exact_state, 0, 1,
         None, .99),
        (h2_hamiltonian, longer_time, h2_initial_state, h2_exact_state, 0, 10,
         LOW_RANK, .9999),
        (lih_hamiltonian, longer_time, lih_initial_state, lih_exact_state, 0, 1,
         LowRankTrotterAlgorithm(final_rank=2), .999),
        (lih_hamiltonian, longer_time, lih_initial_state, lih_exact_state, 0,
         10, LowRankTrotterAlgorithm(final_rank=2), .9999),
    ])
def test_simulate_trotter_simulate(hamiltonian, time, initial_state,
                                   exact_state, order, n_steps, algorithm,
                                   result_fidelity):

    n_qubits = openfermion.count_qubits(hamiltonian)
    qubits = cirq.LineQubit.range(n_qubits)

    start_state = initial_state

    circuit = cirq.Circuit(
        simulate_trotter(qubits, hamiltonian, time, n_steps, order, algorithm))

    final_state = circuit.final_state_vector(initial_state=start_state)
    correct_state = exact_state
    assert fidelity(final_state, correct_state) > result_fidelity
    # Make sure the time wasn't too small
    assert fidelity(final_state, start_state) < 0.95 * result_fidelity


@pytest.mark.parametrize(
    'hamiltonian, time, initial_state, exact_state, order, n_steps, '
    'algorithm, result_fidelity', [
        (diag_coul_hamiltonian, long_time, diag_coul_initial_state,
         diag_coul_exact_state, 0, 5, None, .99),
        (diag_coul_hamiltonian, long_time, diag_coul_initial_state,
         diag_coul_exact_state, 0, 12, None, .999),
        (diag_coul_hamiltonian, long_time, diag_coul_initial_state,
         diag_coul_exact_state, 1, 1, LINEAR_SWAP_NETWORK, .99),
        (diag_coul_hamiltonian, long_time, diag_coul_initial_state,
         diag_coul_exact_state, 2, 1, LINEAR_SWAP_NETWORK, .99999),
        (diag_coul_hamiltonian, long_time, diag_coul_initial_state,
         diag_coul_exact_state, 0, 3, SPLIT_OPERATOR, .99),
        (diag_coul_hamiltonian, long_time, diag_coul_initial_state,
         diag_coul_exact_state, 0, 6, SPLIT_OPERATOR, .999),
        (diag_coul_hamiltonian, long_time, diag_coul_initial_state,
         diag_coul_exact_state, 1, 1, SPLIT_OPERATOR, .99),
        (diag_coul_hamiltonian, long_time, diag_coul_initial_state,
         diag_coul_exact_state, 2, 1, SPLIT_OPERATOR, .99999),
        (h2_hamiltonian, longer_time, h2_initial_state, h2_exact_state, 0, 1,
         LOW_RANK, .99),
        (h2_hamiltonian, longer_time, h2_initial_state, h2_exact_state, 0, 10,
         LOW_RANK, .9999),
        (lih_hamiltonian, longer_time, lih_initial_state, lih_exact_state, 0, 1,
         LowRankTrotterAlgorithm(final_rank=2), .999),
        (lih_hamiltonian, longer_time, lih_initial_state, lih_exact_state, 0,
         10, LowRankTrotterAlgorithm(final_rank=2), .9999),
    ])
def test_simulate_trotter_simulate_controlled(hamiltonian, time, initial_state,
                                              exact_state, order, n_steps,
                                              algorithm, result_fidelity):

    n_qubits = openfermion.count_qubits(hamiltonian)
    qubits = cirq.LineQubit.range(n_qubits)

    control = cirq.LineQubit(-1)
    zero = [1, 0]
    one = [0, 1]
    start_state = (numpy.kron(zero, initial_state) +
                   numpy.kron(one, initial_state)) / numpy.sqrt(2)

    circuit = cirq.Circuit(
        simulate_trotter(qubits, hamiltonian, time, n_steps, order, algorithm,
                         control))

    final_state = circuit.final_state_vector(initial_state=start_state)
    correct_state = (numpy.kron(zero, initial_state) +
                     numpy.kron(one, exact_state)) / numpy.sqrt(2)
    assert fidelity(final_state, correct_state) > result_fidelity
    # Make sure the time wasn't too small
    assert fidelity(final_state, start_state) < 0.95 * result_fidelity


def test_simulate_trotter_omit_final_swaps():
    n_qubits = 5
    qubits = cirq.LineQubit.range(n_qubits)
    hamiltonian = openfermion.DiagonalCoulombHamiltonian(one_body=numpy.ones(
        (n_qubits, n_qubits)),
                                                         two_body=numpy.ones(
                                                             (n_qubits,
                                                              n_qubits)))
    time = 1.0

    circuit_with_swaps = cirq.Circuit(
        simulate_trotter(qubits,
                         hamiltonian,
                         time,
                         order=0,
                         algorithm=LINEAR_SWAP_NETWORK))
    circuit_without_swaps = cirq.Circuit(
        simulate_trotter(qubits,
                         hamiltonian,
                         time,
                         order=0,
                         algorithm=LINEAR_SWAP_NETWORK,
                         omit_final_swaps=True))

    assert len(circuit_without_swaps) < len(circuit_with_swaps)

    circuit_with_swaps = cirq.Circuit(
        simulate_trotter(qubits,
                         hamiltonian,
                         time,
                         order=1,
                         n_steps=3,
                         algorithm=SPLIT_OPERATOR),
        strategy=cirq.InsertStrategy.NEW)
    circuit_without_swaps = cirq.Circuit(simulate_trotter(
        qubits,
        hamiltonian,
        time,
        order=1,
        n_steps=3,
        algorithm=SPLIT_OPERATOR,
        omit_final_swaps=True),
                                         strategy=cirq.InsertStrategy.NEW)

    assert len(circuit_without_swaps) < len(circuit_with_swaps)

    hamiltonian = lih_hamiltonian
    qubits = cirq.LineQubit.range(4)
    circuit_with_swaps = cirq.Circuit(
        simulate_trotter(qubits, hamiltonian, time, order=0,
                         algorithm=LOW_RANK))
    circuit_without_swaps = cirq.Circuit(
        simulate_trotter(qubits,
                         hamiltonian,
                         time,
                         order=0,
                         algorithm=LOW_RANK,
                         omit_final_swaps=True))

    assert len(circuit_without_swaps) < len(circuit_with_swaps)


def test_simulate_trotter_bad_order_raises_error():
    qubits = cirq.LineQubit.range(2)
    hamiltonian = openfermion.random_diagonal_coulomb_hamiltonian(2, seed=0)
    time = 1.0
    with pytest.raises(ValueError):
        _ = next(simulate_trotter(qubits, hamiltonian, time, order=-1))


def test_simulate_trotter_bad_hamiltonian_type_raises_error():
    qubits = cirq.LineQubit.range(2)
    hamiltonian = openfermion.FermionOperator()
    time = 1.0
    with pytest.raises(TypeError):
        _ = next(simulate_trotter(qubits, hamiltonian, time, algorithm=None))
    with pytest.raises(TypeError):
        _ = next(
            simulate_trotter(qubits,
                             hamiltonian,
                             time,
                             algorithm=LINEAR_SWAP_NETWORK))


def test_simulate_trotter_unsupported_trotter_step_raises_error():
    qubits = cirq.LineQubit.range(2)
    control = cirq.LineQubit(-1)
    hamiltonian = openfermion.random_diagonal_coulomb_hamiltonian(2, seed=0)
    time = 1.0

    class EmptyTrotterAlgorithm(TrotterAlgorithm):
        supported_types = {openfermion.DiagonalCoulombHamiltonian}

    algorithm = EmptyTrotterAlgorithm()
    with pytest.raises(ValueError):
        _ = next(
            simulate_trotter(qubits,
                             hamiltonian,
                             time,
                             order=0,
                             algorithm=algorithm))
    with pytest.raises(ValueError):
        _ = next(
            simulate_trotter(qubits,
                             hamiltonian,
                             time,
                             order=1,
                             algorithm=algorithm))
    with pytest.raises(ValueError):
        _ = next(
            simulate_trotter(qubits,
                             hamiltonian,
                             time,
                             order=0,
                             algorithm=algorithm,
                             control_qubit=control))
    with pytest.raises(ValueError):
        _ = next(
            simulate_trotter(qubits,
                             hamiltonian,
                             time,
                             order=1,
                             algorithm=algorithm,
                             control_qubit=control))


@pytest.mark.parametrize('algorithm_type,hamiltonian', [
    (LINEAR_SWAP_NETWORK, openfermion.random_diagonal_coulomb_hamiltonian(2)),
    (LOW_RANK, openfermion.random_interaction_operator(2)),
    (SPLIT_OPERATOR, openfermion.random_diagonal_coulomb_hamiltonian(2)),
])
def test_trotter_misspecified_control_raises_error(algorithm_type, hamiltonian):
    qubits = cirq.LineQubit.range(2)
    time = 2.

    algorithms = [
        algorithm_type.controlled_asymmetric(hamiltonian),
        algorithm_type.controlled_symmetric(hamiltonian)
    ]

    for algorithm in algorithms:
        if algorithm is None:
            continue
        with pytest.raises(TypeError):
            next(algorithm.trotter_step(qubits, time))
        with pytest.raises(TypeError):
            next(algorithm.trotter_step(qubits, time, control_qubit=2))



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Circuit generation functions for verified phase estimation (2010.02538)"""

from typing import Sequence, Optional
import numpy
import cirq


def vpe_single_circuit(qubits: Sequence[cirq.Qid], prep: cirq.Circuit,
                       evolve: cirq.Circuit, initial_rotation: cirq.Gate,
                       final_rotation: cirq.Gate) -> cirq.Circuit:
    """
    Combines the different parts that make up a VPE circuit

    The protocol for VPE requires combining preparation, evolution, and
    measurement circuits for different values of time in order to estimate
    the phase function. This function takes these parts and combines them.

    Note that we need not specify the time of evolution as this is contained
    already within evolve.

    Arguments:
        prep [cirq.Circuit] -- The circuit to prepare the initial state
            (|psi_s>+|psi_r>) from |0>+|1>
        evolve [cirq.Circuit] -- The circuit to evolve for time t
        initial_rotation [cirq.Gate] -- The initial rotation on the target qubit
            (Note that the gate should already be targeting the qubit)
        final_rotation [cirq.Gate] -- The final rotation on the target qubit
            (Note that the gate should already be targeting the qubit)
    """
    circuit = cirq.Circuit()
    circuit.append(initial_rotation)
    circuit.append(prep)
    circuit.append(evolve)
    circuit.append(cirq.inverse(prep))
    circuit.append(final_rotation)
    circuit.append(cirq.measure(*qubits, key='msmt'))
    return circuit


# Turning off yapf here as its formatting suggestion is bad.
# yapf: disable
standard_vpe_rotation_set = [
    [0.25, cirq.ry(numpy.pi / 2), cirq.ry(-numpy.pi / 2)],
    [-0.25, cirq.ry(numpy.pi / 2), cirq.ry(numpy.pi / 2)],
    [-0.25j, cirq.ry(numpy.pi / 2), cirq.rx(-numpy.pi / 2)],
    [0.25j, cirq.ry(numpy.pi / 2), cirq.rx(numpy.pi / 2)],
    [0.25, cirq.rx(numpy.pi / 2), cirq.rx(-numpy.pi / 2)],
    [-0.25, cirq.rx(numpy.pi / 2), cirq.rx(numpy.pi / 2)],
    [0.25j, cirq.rx(numpy.pi / 2), cirq.ry(-numpy.pi / 2)],
    [-0.25j, cirq.rx(numpy.pi / 2), cirq.ry(numpy.pi / 2)],
]
# yapf: enable


def vpe_circuits_single_timestep(qubits: Sequence[cirq.Qid],
                                 prep: cirq.Circuit,
                                 evolve: cirq.Circuit,
                                 target_qubit: cirq.Qid,
                                 rotation_set: Optional[Sequence] = None
                                ) -> Sequence[cirq.Circuit]:
    """Prepares the circuits to perform VPE at a fixed time

    Puts together the set of pre- and post-rotations to implement
    VPE at for a given state preparation and time evolution.

    [description]

    Arguments:
        prep [cirq.Circuit] -- The circuit to prepare the target state
            (|psi_s>+|psi_r>) from |0>+|1>
        evolve [cirq.Circuit] -- The circuit to evolve for time t
        target_qubit [cirq.Qid] -- The qubit on which the phase
            function is encoded
        rotation_set [Sequence] -- A set of initial and final rotations for the
            target qubit. We average the phase function estimation over multiple
            such rotations to cancel out readout noise, final T1 decay, etc.
            The standard rotation set is typically sufficient for these
            purposes. The first element of each gate is the multiplier to get
            the phase function; we do not need this for this function.

            If rotation_set is set to None, the 'standard rotation set' of all
            possible X and Y rotations before and after the circuit is used.
    """
    if rotation_set is None:
        rotation_set = standard_vpe_rotation_set
    circuits = [
        vpe_single_circuit(qubits, prep, evolve, rdata[1].on(target_qubit),
                           rdata[2].on(target_qubit)) for rdata in rotation_set
    ]
    return circuits



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Test for qubit_tapering_from_stabilizer model."""

import unittest

import numpy

from openfermion.chem import MolecularData
from openfermion.ops.operators import QubitOperator
from openfermion.transforms.opconversions import (jordan_wigner,
                                                  get_fermion_operator)
from openfermion.linalg import eigenspectrum
from openfermion.utils.operator_utils import count_qubits

from openfermion.transforms.repconversions.qubit_tapering_from_stabilizer \
    import (StabilizerError, check_commuting_stabilizers,
            check_stabilizer_linearity, fix_single_term, _reduce_terms,
            _reduce_terms_keep_length, _lookup_term, reduce_number_of_terms,
            taper_off_qubits)


def lih_hamiltonian():
    """
    Generate test Hamiltonian from LiH.

    Args:
        None

    Return:

        hamiltonian: FermionicOperator

        spectrum: List of energies.
    """
    geometry = [('Li', (0., 0., 0.)), ('H', (0., 0., 1.45))]
    active_space_start = 1
    active_space_stop = 3
    molecule = MolecularData(geometry, 'sto-3g', 1, description="1.45")
    molecule.load()

    molecular_hamiltonian = molecule.get_molecular_hamiltonian(
        occupied_indices=range(active_space_start),
        active_indices=range(active_space_start, active_space_stop))

    hamiltonian = get_fermion_operator(molecular_hamiltonian)
    spectrum = eigenspectrum(hamiltonian)

    return hamiltonian, spectrum


class TaperingTest(unittest.TestCase):
    """TaperingTest class."""

    def test_function_errors(self):
        """Test error of main function."""
        hamiltonian, _ = lih_hamiltonian()
        qubit_hamiltonian = jordan_wigner(hamiltonian)
        stab1 = QubitOperator('Z0 Z2', -1.0)
        stab2 = QubitOperator('Z1 Z3', -1.0)
        with self.assertRaises(TypeError):
            reduce_number_of_terms(operator=1, stabilizers=stab1 + stab2)
        with self.assertRaises(TypeError):
            reduce_number_of_terms(operator=qubit_hamiltonian, stabilizers=1)
        with self.assertRaises(TypeError):
            reduce_number_of_terms(operator=qubit_hamiltonian,
                                   stabilizers=stab1 + stab2,
                                   manual_input=True,
                                   fixed_positions=None)
        with self.assertRaises(StabilizerError):
            reduce_number_of_terms(operator=qubit_hamiltonian,
                                   stabilizers=stab1 + stab2,
                                   manual_input=True,
                                   fixed_positions=[1])
        with self.assertRaises(StabilizerError):
            reduce_number_of_terms(operator=qubit_hamiltonian,
                                   stabilizers=stab1 + stab2,
                                   manual_input=True,
                                   fixed_positions=[1, 1])
        with self.assertRaises(StabilizerError):
            # Check Identity as stabilizer error.
            reduce_number_of_terms(operator=qubit_hamiltonian,
                                   stabilizers=(stab1 +
                                                QubitOperator(' ', 1.0)))
        with self.assertRaises(StabilizerError):
            # Check complex coefficient stabilizer error.
            reduce_number_of_terms(operator=qubit_hamiltonian,
                                   stabilizers=(stab1 +
                                                QubitOperator('Z0', 1.0j)))
        with self.assertRaises(StabilizerError):
            # Check linearly-dependent stabilizer error.
            reduce_number_of_terms(
                operator=qubit_hamiltonian,
                stabilizers=(stab1 + QubitOperator('Z0 Z1 Z2 Z3', 1.0) + stab2))
        with self.assertRaises(StabilizerError):
            # Check anti-commuting stabilizer error.
            reduce_number_of_terms(operator=qubit_hamiltonian,
                                   stabilizers=(QubitOperator('X0', 1.0) +
                                                QubitOperator('Y0', 1.0)))
        with self.assertRaises(StabilizerError):
            # Check linearly-dependent stabilizer error.
            _reduce_terms(
                terms=qubit_hamiltonian,
                stabilizer_list=list(stab1 + QubitOperator('Z0 Z1 Z2 Z3', 1.0) +
                                     stab2),
                manual_input=False,
                fixed_positions=[])
        with self.assertRaises(StabilizerError):
            # Check complex coefficient stabilizer error.
            _reduce_terms(terms=qubit_hamiltonian,
                          stabilizer_list=list(stab1 +
                                               QubitOperator('Z0', 1.0j)),
                          manual_input=False,
                          fixed_positions=[])
        with self.assertRaises(StabilizerError):
            # Check linearly-dependent stabilizer error.
            par_qop = QubitOperator('Z0 Z1 Z2 Z3', 1.0)
            _reduce_terms_keep_length(terms=qubit_hamiltonian,
                                      stabilizer_list=[stab1, par_qop, stab2],
                                      manual_input=False,
                                      fixed_positions=[])
        with self.assertRaises(StabilizerError):
            # Check complex coefficient stabilizer error.
            aux_qop = QubitOperator('Z0', 1.0j)
            _reduce_terms_keep_length(terms=qubit_hamiltonian,
                                      stabilizer_list=[stab1, aux_qop],
                                      manual_input=False,
                                      fixed_positions=[])
        with self.assertRaises(StabilizerError):
            # Test check_commuting_stabilizer function
            # Requires a list of QubitOperators one of which
            # has an imaginary term.
            check_commuting_stabilizers(stabilizer_list=[
                QubitOperator('Z0 Z1', 1.0),
                QubitOperator('X0', 1j)
            ],
                                        msg='This test fails.')
        with self.assertRaises(StabilizerError):
            # Test check_stabilizer_linearity function.
            # Requires a list of QUbitOperators one of which is
            # the identity.
            check_stabilizer_linearity(
                [QubitOperator('Z0 Z1', 1.0),
                 QubitOperator(' ', 1.0)],
                msg='This test fails.')

    def test_fix_single_term(self):
        """Test fix_single_term function."""
        stab2 = QubitOperator('Z1 Z3', -1.0)
        test_term = QubitOperator('Z1 Z2')

        fix1 = fix_single_term(test_term, 1, 'Z', 'X', stab2)
        fix2 = fix_single_term(test_term, 0, 'X', 'X', stab2)

        self.assertTrue(fix1 == (test_term * stab2))
        self.assertTrue(fix2 == test_term)

    def test_lookup_term(self):
        """Test for the auxiliar function _lookup_term."""
        # Dummy test where the initial Pauli string is larger than the
        # updated one.
        start_op = list(QubitOperator('Z0 Z1 Z2 Z3').terms.keys())[0]
        updateop1 = QubitOperator('Z0 Z2', -1.0)
        updateop2 = list(QubitOperator('Z0 Z1 Z2 Z3').terms.keys())

        qop = _lookup_term(start_op, [updateop1], updateop2)
        final_op = list(qop.terms.keys())[0]

        self.assertLess(len(final_op), len(start_op))

    def test_reduce_terms(self):
        """Test reduce_terms function using LiH Hamiltonian."""
        hamiltonian, spectrum = lih_hamiltonian()
        qubit_hamiltonian = jordan_wigner(hamiltonian)
        stab1 = QubitOperator('Z0 Z2', -1.0)
        stab2 = QubitOperator('Z1 Z3', -1.0)

        red_eigenspectrum = eigenspectrum(
            reduce_number_of_terms(qubit_hamiltonian, stab1 + stab2))

        self.assertAlmostEqual(spectrum[0], red_eigenspectrum[0])

    def test_reduce_terms_manual_input(self):
        """Test reduce_terms function using LiH Hamiltonian."""
        hamiltonian, spectrum = lih_hamiltonian()
        qubit_hamiltonian = jordan_wigner(hamiltonian)
        stab1 = QubitOperator('Z0 Z2', -1.0)
        stab2 = QubitOperator('Z1 Z3', -1.0)

        red_eigenspectrum = eigenspectrum(
            reduce_number_of_terms(qubit_hamiltonian, [stab1, stab2],
                                   manual_input=True,
                                   fixed_positions=[0, 1]))

        self.assertAlmostEqual(spectrum[0], red_eigenspectrum[0])

    def test_reduce_terms_maintain_length(self):
        """Test reduce_terms function using LiH Hamiltonian."""
        hamiltonian, spectrum = lih_hamiltonian()
        qubit_hamiltonian = jordan_wigner(hamiltonian)
        stab1 = QubitOperator('Z0 Z2', -1.0)
        stab2 = QubitOperator('Z1 Z3', -1.0)

        red_eigenspectrum = eigenspectrum(
            reduce_number_of_terms(qubit_hamiltonian,
                                   stab1 + stab2,
                                   maintain_length=True))

        self.assertAlmostEqual(spectrum[0], red_eigenspectrum[0])

    def test_reduce_terms_auxiliar_functions(self):
        """Test reduce_terms function using LiH Hamiltonian."""
        hamiltonian, spectrum = lih_hamiltonian()
        qubit_ham = jordan_wigner(hamiltonian)
        stab1 = QubitOperator('Z0 Z2', -1.0)
        stab2 = QubitOperator('Z1 Z3', -1.0)

        red_ham1, _ = _reduce_terms(terms=qubit_ham,
                                    stabilizer_list=[stab1, stab2],
                                    manual_input=False,
                                    fixed_positions=[])
        red_ham2, _ = _reduce_terms_keep_length(terms=qubit_ham,
                                                stabilizer_list=[stab1, stab2],
                                                manual_input=False,
                                                fixed_positions=[])
        red_eigspct1 = eigenspectrum(red_ham1)
        red_eigspct2 = eigenspectrum(red_ham2)

        self.assertAlmostEqual(spectrum[0], red_eigspct1[0])
        self.assertAlmostEqual(spectrum[0], red_eigspct2[0])

    def test_reduce_terms_auxiliar_functions_manual_input(self):
        """Test reduce_terms function using LiH Hamiltonian."""
        hamiltonian, spectrum = lih_hamiltonian()
        qubit_ham = jordan_wigner(hamiltonian)
        stab1 = QubitOperator('Z0 Z2', -1.0)
        stab2 = QubitOperator('Z1 Z3', -1.0)

        red_ham1, _ = _reduce_terms(terms=qubit_ham,
                                    stabilizer_list=[stab1, stab2],
                                    manual_input=True,
                                    fixed_positions=[0, 1])
        red_ham2, _ = _reduce_terms_keep_length(terms=qubit_ham,
                                                stabilizer_list=[stab1, stab2],
                                                manual_input=True,
                                                fixed_positions=[0, 1])
        red_eigspct1 = eigenspectrum(red_ham1)
        red_eigspct2 = eigenspectrum(red_ham2)

        self.assertAlmostEqual(spectrum[0], red_eigspct1[0])
        self.assertAlmostEqual(spectrum[0], red_eigspct2[0])

    def test_tapering_qubits_manual_input_false(self):
        """Test taper_off_qubits function using LiH Hamiltonian."""
        hamiltonian, spectrum = lih_hamiltonian()
        qubit_hamiltonian = jordan_wigner(hamiltonian)
        stab1 = QubitOperator('Z0 Z2', -1.0)
        stab2 = QubitOperator('Z1 Z3', -1.0)

        tapered_hamiltonian = taper_off_qubits(operator=qubit_hamiltonian,
                                               stabilizers=[stab1, stab2],
                                               manual_input=False,
                                               fixed_positions=[0, 3])
        tapered_spectrum = eigenspectrum(tapered_hamiltonian)

        self.assertAlmostEqual(spectrum[0], tapered_spectrum[0])

    def test_tapering_qubits_manual_input(self):
        """
        Test taper_off_qubits function using LiH Hamiltonian.

        Checks different qubits inputs to remove manually.

        Test the lowest eigenvalue against the full Hamiltonian,
        and the full spectrum between them.
        """
        hamiltonian, spectrum = lih_hamiltonian()
        qubit_hamiltonian = jordan_wigner(hamiltonian)
        stab1 = QubitOperator('Z0 Z2', -1.0)
        stab2 = QubitOperator('Z1 Z3', -1.0)

        tapered_ham_0_3 = taper_off_qubits(qubit_hamiltonian, [stab1, stab2],
                                           manual_input=True,
                                           fixed_positions=[0, 3])
        tapered_ham_2_1 = taper_off_qubits(qubit_hamiltonian, [stab1, stab2],
                                           manual_input=True,
                                           fixed_positions=[2, 1])

        tapered_spectrum_0_3 = eigenspectrum(tapered_ham_0_3)
        tapered_spectrum_2_1 = eigenspectrum(tapered_ham_2_1)

        self.assertAlmostEqual(spectrum[0], tapered_spectrum_0_3[0])
        self.assertAlmostEqual(spectrum[0], tapered_spectrum_2_1[0])
        self.assertTrue(
            numpy.allclose(tapered_spectrum_0_3, tapered_spectrum_2_1))

    def test_tapering_qubits_remove_positions(self):
        """Test taper_off_qubits function using LiH Hamiltonian."""
        hamiltonian, spectrum = lih_hamiltonian()
        qubit_hamiltonian = jordan_wigner(hamiltonian)
        stab1 = QubitOperator('Z0 Z2', -1.0)
        stab2 = QubitOperator('Z1 Z3', -1.0)

        (tapered_hamiltonian,
         positions) = taper_off_qubits(operator=qubit_hamiltonian,
                                       stabilizers=[stab1, stab2],
                                       manual_input=True,
                                       fixed_positions=[0, 3],
                                       output_tapered_positions=True)

        tapered_spectrum = eigenspectrum(tapered_hamiltonian)

        self.assertAlmostEqual(spectrum[0], tapered_spectrum[0])
        self.assertEqual(positions, [0, 3])

    def test_tappering_stabilizer_more_qubits(self):
        """Test for stabilizer with more qubits than operator."""
        hamiltonian = QubitOperator('Y0 Y1', 1.0)
        stab = QubitOperator('X0 X1 X2', -1.0)

        num_qubits = max(count_qubits(hamiltonian), count_qubits(stab))
        tap_ham = taper_off_qubits(hamiltonian, stab)
        num_qubits_tap = count_qubits(tap_ham)

        self.assertFalse(num_qubits == num_qubits_tap)

        hamiltonian = QubitOperator('X0 X1', 1.0)
        stab = QubitOperator('Y0 Y1 Y2', -1.0)

        num_qubits = max(count_qubits(hamiltonian), count_qubits(stab))
        tap_ham = taper_off_qubits(hamiltonian, stab)
        num_qubits_tap = count_qubits(tap_ham)
        self.assertFalse(num_qubits == num_qubits_tap)

    def test_taper_x_stab(self):
        hamiltonian = QubitOperator('X0 X1', 1.0)
        stab1 = QubitOperator('Y0 Y2', -1.0)

        tham = reduce_number_of_terms(hamiltonian, stab1, maintain_length=True)
        self.assertTrue(hamiltonian == tham)


import numpy as np
import pytest
import scipy as sp
from openfermion.third_party.representability._namedtensor import Tensor
from openfermion.third_party.representability._multitensor import MultiTensor, \
    TMap
from openfermion.third_party.representability._dualbasis import DualBasis, \
    DualBasisElement


def test_tmap():
    a = np.random.random((5, 5))
    b = np.random.random((4, 4))
    c = np.random.random((3, 3))
    at = Tensor(tensor=a, name='a')
    bt = Tensor(tensor=b, name='b')
    ct = Tensor(tensor=c, name='c')
    ttmap = TMap(tensors=[at, bt, ct])
    assert np.allclose(ttmap['a'].data, a)
    assert np.allclose(ttmap['b'].data, b)
    assert np.allclose(ttmap['c'].data, c)

    tmp_tensors = [a, b, c]
    for idx, iterated_tensor in enumerate(ttmap):
        assert np.allclose(iterated_tensor.data, tmp_tensors[idx])


def test_multitensor_init():
    """
    Testing the generation of a multitensor object with random tensors
    """
    a = np.random.random((5, 5))
    b = np.random.random((4, 4))
    c = np.random.random((3, 3))
    at = Tensor(tensor=a, name='a')
    bt = Tensor(tensor=b, name='b')
    ct = Tensor(tensor=c, name='c')
    mt = MultiTensor([at, bt, ct])

    with pytest.raises(TypeError):
        _ = MultiTensor((at, bt))

    assert len(mt.dual_basis) == 0

    assert np.isclose(mt.vec_dim, 5**2 + 4**2 + 3**2)


def test_multitensor_offsetmap():
    a = np.random.random((5, 5, 5, 5))
    b = np.random.random((4, 4, 4))
    c = np.random.random((3, 3))
    at = Tensor(tensor=a, name='a')
    bt = Tensor(tensor=b, name='b')
    ct = Tensor(tensor=c, name='c')
    mt = MultiTensor([at, bt, ct])

    assert mt.off_set_map == {'a': 0, 'b': 5**4, 'c': 5**4 + 4**3}


def test_vectorize_test():
    a = np.random.random((5, 5))
    b = np.random.random((4, 4))
    c = np.random.random((3, 3))
    at = Tensor(tensor=a, name='a')
    bt = Tensor(tensor=b, name='b')
    ct = Tensor(tensor=c, name='c')
    mt = MultiTensor([at, bt, ct])
    vec = np.vstack((at.vectorize(), bt.vectorize()))
    vec = np.vstack((vec, ct.vectorize()))
    assert np.allclose(vec, mt.vectorize_tensors())

    a = np.random.random((5, 5, 5, 5))
    b = np.random.random((4, 4, 4))
    c = np.random.random((3, 3))
    at = Tensor(tensor=a, name='a')
    bt = Tensor(tensor=b, name='b')
    ct = Tensor(tensor=c, name='c')
    mt = MultiTensor([at, bt, ct])
    vec = np.vstack((at.vectorize(), bt.vectorize()))
    vec = np.vstack((vec, ct.vectorize()))
    assert np.allclose(vec, mt.vectorize_tensors())


def test_add_dualelement():
    a = np.random.random((5, 5, 5, 5))
    b = np.random.random((4, 4, 4))
    c = np.random.random((3, 3))
    at = Tensor(tensor=a, name='a')
    bt = Tensor(tensor=b, name='b')
    ct = Tensor(tensor=c, name='c')
    mt = MultiTensor([at, bt, ct])
    assert isinstance(mt.dual_basis, DualBasis)

    dbe = DualBasisElement()
    dbe.add_element('a', (0, 1, 2, 3), 4)
    mt.add_dual_elements(dbe)
    assert len(mt.dual_basis) == 1


def test_synthesis_element():
    a = np.random.random((5, 5))
    b = np.random.random((4, 4))
    c = np.random.random((3, 3))
    at = Tensor(tensor=a, name='a')
    bt = Tensor(tensor=b, name='b')
    ct = Tensor(tensor=c, name='c')
    mt = MultiTensor([at, bt, ct])

    dbe = DualBasisElement()
    dbe.add_element('a', (0, 1), 4)
    dbe.add_element('a', (1, 0), 4)
    with pytest.raises(TypeError):
        dbe.add_element(5)

    with pytest.raises(TypeError):
        mt.add_dual_elements(5)

    mt.add_dual_elements(dbe)
    colidx, data_vals = mt.synthesize_element(dbe)
    assert data_vals == [4, 4]
    assert colidx == [1, 5]
    assert [at.data[0, 1], at.data[1, 0]] == [at(0, 1), at(1, 0)]


def test_synthesis_dualbasis():
    a = np.random.random((5, 5))
    b = np.random.random((4, 4))
    c = np.random.random((3, 3))
    at = Tensor(tensor=a, name='a')
    bt = Tensor(tensor=b, name='b')
    ct = Tensor(tensor=c, name='c')
    dbe = DualBasisElement()
    dbe.add_element('a', (0, 1), 4)
    dbe.add_element('a', (1, 0), 4)
    mt = MultiTensor([at, bt, ct], DualBasis(elements=[dbe]))

    A, c, b = mt.synthesize_dual_basis()
    assert isinstance(A, sp.sparse.csr_matrix)
    assert isinstance(c, sp.sparse.csr_matrix)
    assert isinstance(b, sp.sparse.csr_matrix)
    assert A.shape == (1, 50)
    assert b.shape == (1, 1)
    assert c.shape == (1, 1)


def test_dual_basis_element():
    de = DualBasisElement()
    de_2 = DualBasisElement()
    db_0 = de + de_2
    assert isinstance(db_0, DualBasis)
    db_1 = db_0 + db_0
    assert isinstance(db_1, DualBasis)

    dim = 2
    opdm = np.random.random((dim, dim))
    opdm = (opdm.T + opdm) / 2
    opdm = Tensor(tensor=opdm, name='opdm')
    rdm = MultiTensor([opdm])

    def generate_dual_basis_element(i, j):
        element = DualBasisElement(tensor_names=["opdm"],
                                   tensor_elements=[(i, j)],
                                   tensor_coeffs=[-1.0],
                                   bias=1 if i == j else 0,
                                   scalar=0)
        return element

    opdm_to_oqdm_map = DualBasis()
    for _, idx in opdm.all_iterator():
        i, j = idx
        opdm_to_oqdm_map += generate_dual_basis_element(i, j)

    rdm.dual_basis = opdm_to_oqdm_map
    A, b, _ = rdm.synthesize_dual_basis()
    Adense = A.todense()
    opdm_flat = opdm.data.reshape((-1, 1))
    oqdm = Adense.dot(opdm_flat)
    test_oqdm = oqdm + b.todense()
    assert np.allclose(test_oqdm.reshape((dim, dim)), np.eye(dim) - opdm.data)


def test_cover_make_offset_dict():
    a = np.random.random((5, 5))
    b = np.random.random((4, 4))
    c = np.random.random((3, 3))
    with pytest.raises(TypeError):
        _ = MultiTensor.make_offset_dict([a, b, c])
